# Spatial Correlograms
from collections.abc import Callable

import geopandas as gpd
import pandas as pd
import numpy as np
from joblib import Parallel, delayed
from libpysal.cg.kdtree import KDTree
from libpysal.weights import KNN, DistanceBand
from libpysal.weights.util import get_points_array
from sklearn.metrics import pairwise_distances
from scipy import spatial, linalg
import warnings
from .moran import Moran


def _get_stat(inputs: tuple) -> pd.Series:
    """helper function for computing parallel statistics at multiple Graph specifications

    Parameters
    ----------
    inputs : tuple
        tuple of (y, tree, W, statistic, STATISTIC, dist, weights_kwargs, stat_kwargs)

    Returns
    -------
    pandas.Series
        a pandas series with the computed autocorrelation statistic and its simulated p-value
    """
    (
        y,  # y variable
        tree,  # kdreee,
        W,  # weights class
        STATISTIC,  # class of statistic (Moran, Geary, etc)
        dist,  # threshold/k parameter for the weights
        weights_kwargs,  # additional args
        stat_kwargs,  # additional args
    ) = inputs

    w = W(tree, dist, silence_warnings=True, **weights_kwargs)
    with warnings.catch_warnings(action="ignore", category=RuntimeWarning):
        autocorr = STATISTIC(y, w, **stat_kwargs)
    attrs = []
    all_attrs = list(dict(vars(autocorr)).keys())
    for attribute in all_attrs:
        attrs.append(getattr(autocorr, str(attribute)))
    return pd.Series(attrs, index=all_attrs, name=dist)


def correlogram(
    geometry: gpd.GeoSeries,
    variable: str | list | pd.Series | None,
    support: list | None = None,
    statistic: Callable | str = Moran,
    distance_type: str = "band",
    weights_kwargs: dict = None,
    stat_kwargs: dict = None,
    select_numeric: bool = False,
    n_jobs: int = -1,
    n_bins: int | None = 50,
) -> pd.DataFrame:
    """Generate a spatial correlogram

    A spatial profile is a set of spatial autocorrelation statistics calculated for
    a set of increasing distances. It is a useful exploratory tool for examining
    how the relationship between spatial units changes over different notions of scale.

    Parameters
    ----------
    geometry : gpd.GeoSeries
        geodataframe holding spatial and attribute data
    variable: pd.Series or list
        pandas series matching input geometries
    support : list or None
        list of values at which to compute the autocorrelation statistic
    statistic : callable or str
        statistic to be computed for a range of libpysal.Graph specifications.
        This should be a class with a signature like ``Statistic(y,w, **kwargs)``
        where ``y`` is a  array and ``w`` is a ``libpysal.weights.W`` object
        Generally, this is a class from pysal's ``esda`` package
        defaults to ``esda.Moran``, which computes the Moran's I statistic. If
        ``'lowess'`` is provided, a non-parametric correlogram is computed using
        lowess regression on the spatial-covariation model, see Notes.
    distance_type : str, optional
        which concept of distance to increment. Options are ``{`band`, `knn`}``.
        by default ``'band'`` (for ``libpysal.weights.DistanceBand`` weights)
    weights_kwargs : dict
        additional keyword arguments passed to the ``libpysal.weights.W`` class
    stat_kwargs : dict
        additional keyword arguments passed to the ``esda`` autocorrelation statistic class.
        For example for faster results with no statistical inference, set the number
        of permutations to zero with ``stat_kwargs={permutations: 0}``
    select_numeric : bool
        if True, only return numeric attributes from the original class. This is useful
        e.g. to prevent lists inside a "cell" of a dataframe
    n_jobs : int
        number of jobs to pass to joblib. If -1 (default), all cores will be used
    n_bins : int
        number of distance bands or k-nearest neighbor values to use if
        ``support`` is not provided. Ignored if ``support`` is provided.
        by default 10. If ``distance_type`` is 'knn', the number of neighbors
        will be capped at n-1, where n is the number of observations. Further,
        if n-1 is not divisible by ``n_bins``, the actual number of bins will be
        may be off by one bin.

    Returns
    -------
    outputs : pandas.DataFrame
        table of autocorrelation statistics at increasing distance bandwidths

    Notes
    -----
    The nonparametric correlogram uses a lowess regression
    to estimate the spatial-covariation model:

    .. math::

        zi*zj = f(d_{ij}) + e_ij

    where :math:`f` is a smooth function of distance :math:`d_{ij}` between points
    :math:`i` and :math:`j`. This function requires the statsmodels package to be
    installed.

    For the nonparametric correlogram, a precomputed distance matrix can
    be used. To do this, set
    ``stat_kwargs={'metric':'precomputed', 'coordinates':distance_matrix}``
    where ``distance_matrix`` is a square matrix of pairwise distances that
    aligns with the ``geometry`` rows.
    """
    if stat_kwargs is None:
        stat_kwargs = dict()
    if weights_kwargs is None:
        weights_kwargs = dict()

    if isinstance(geometry, gpd.GeoDataFrame):
        geometry = geometry.geometry
    elif not isinstance(geometry, gpd.GeoSeries):
        raise ValueError("geometry must be a geopandas GeoDataFrame or GeoSeries")

    if not (geometry.type == "Point").all():
        raise ValueError(
            "geometry must be of type Point. Try sending geometry.centroid"
        )

    tree = KDTree(get_points_array(geometry.values))

    if support is None:
        if distance_type == "band":
            stop = (
                spatial.distance.cdist(
                    tree.maxes.reshape(1, 2), tree.mins.reshape(1, 2), "euclidean"
                ).item()
                / 2
            )
            d, i = tree.query(tree.data, k=2)
            start = d[d > 0].min()
            support = np.linspace(start, stop, n_bins).squeeze().tolist()
        else:
            n_samples = geometry.shape[0]
            step = (n_samples - 1) / (n_bins - 1)
            # not guaranteed to be n_bins if n_samples not divisible by n_bins
            support = [*np.arange(1, n_samples - 1, step).astype(int), n_samples - 1]
    if distance_type == "band":
        W = DistanceBand
    elif distance_type == "knn":
        if max(support) > (geometry.shape[0] - 1):
            raise ValueError(
                "max number of neighbors must be less than or equal to n-1"
            )
        W = KNN
    else:
        raise ValueError("distance_type must be either `band` or `knn`")

    y = np.asarray(variable).squeeze()

    if y.shape[0] != geometry.shape[0]:
        raise ValueError(
            f"variable is length {len(y)} but geometry has {geometry.shape[0]} rows"
        )

    if statistic != "lowess":
        inputs = [
            (
                y,
                tree,
                W,
                statistic,
                dist,
                weights_kwargs,
                stat_kwargs,
            )
            for dist in support
        ]

        outputs = Parallel(n_jobs=n_jobs, backend="loky")(
            delayed(_get_stat)(i) for i in inputs
        )
    elif statistic == "lowess":
        # lowess correlogram
        stat_kwargs.setdefault("coordinates", tree.data)
        outputs = _lowess_correlogram(y, xvals=support, **stat_kwargs)
    else:
        raise ValueError(
            f"statistic must be a callable or 'lowess', recieved {statistic}"
        )

    df = pd.DataFrame(outputs)
    if select_numeric:
        df = df.select_dtypes(["number"])
    return df


def _lowess_correlogram(
    y: np.ndarray,
    coordinates: np.ndarray,
    xvals: np.ndarray,
    metric: str = "euclidean",
    **lowess_args,
) -> pd.DataFrame:
    """
    Compute a nonparametric correlogram using a kernel regression
    on the spatial-covariation model:

        zi*zj = f(d_{ij}) + e_ij

    where f is a smooth function of distance d_{ij} between points i and j.

    Arguments
    ---------
    y : array-like
        1D array of values to compute the correlogram on
    coordinates : array-like
        2D array of point coordinates or a precomputed distance matrix
    xvals : array-like
        1D array of distance values to evaluate the correlogram at
    metric : str
        distance metric to use. Any metric from sklearn.metrics.pairwise_distances
        is allowed. If 'precomputed', then coordinates is assumed to be a distance matrix
    lowess_args : keyword arguments
        additional keyword arguments passed to statsmodels.nonparametric.smoothers_lowess.lowess

    Returns
    -------
    pandas.DataFrame
        dataframe with index of xvals and a single column 'lowess' with the smoothed
        correlogram values

    Notes
    -----
    This function requires the statsmodels package to be installed. Further, no
    validation is done on the input parameters.
    """
    try:
        from statsmodels.nonparametric.smoothers_lowess import lowess
    except ImportError as e:
        raise ImportError("Nonparametric correlograms require statsmodels") from e

    if metric == "precomputed":
        d = coordinates  # assume this is a distance matrix
    else:
        d = pairwise_distances(coordinates, metric=metric)

    z = (y - y.mean()) / y.std()
    cov = np.multiply.outer(z, z)

    # this assumes that xvals are sorted and span the entire range of distances
    # this is often not the case, so we need to calculate what the actual bin width is.

    xvals = np.asarray(xvals)
    xvals.sort()
    n_samples = d.shape[0]
    if len(xvals) == 1:
        bin_frac = 1.0
    else:
        if len(xvals) == 2:
            lo_width = hi_width = xvals[1] - xvals[0]
        else:  # only one xval, so just use a default width
            lo_width = xvals[1] - xvals[0]
            hi_width = xvals[-1] - xvals[-2]
        lo = max(xvals[0] - lo_width / 2, 0)  # clip to zero
        hi = xvals[-1] + hi_width / 2
        # fraction of off-diagonal values spanned by bins, handling co-location
        frac_in_range = (d[(d >= lo) & (d <= hi)].size - n_samples) / (
            n_samples * (n_samples - 1)
        )
        bin_frac = frac_in_range / len(xvals)

    lowess_args.setdefault("frac", bin_frac)

    if metric != "precomputed":
        row, col = np.triu_indices_from(cov, k=1)
        smooth = lowess(cov[row, col], d[row, col], xvals=xvals, **lowess_args)
    else:  # can't use upper triangle if d is not symmetric
        if linalg.issymmetric(d):
            row, col = np.triu_indices_from(cov)
            smooth = lowess(cov[row, col], d[row, col], xvals=xvals, **lowess_args)
        else:
            smooth = lowess(
                endog=cov.flatten(), exog=d.flatten(), xvals=xvals, **lowess_args
            )

    return pd.DataFrame(smooth, index=xvals, columns=["lowess"])



import numpy as np

from .. import mixture_smoothing as m_s


class TestNPMixtureSmoother:
    """Mixture Smoothing Unit Tests"""

    def setup_method(self):
        self.e = np.array([10, 5, 12, 20])
        self.b = np.array([100, 150, 80, 200])

    def test_defaults(self):
        """Test the main class"""
        mix = m_s.NP_Mixture_Smoother(self.e, self.b)

        np.testing.assert_array_almost_equal(
            mix.r, np.array([0.10982278, 0.03445531, 0.11018404, 0.11018604])
        )
        np.testing.assert_array_almost_equal(mix.category, np.array([1, 0, 1, 1]))

        left, right = mix.getSeed()
        np.testing.assert_array_almost_equal(left, np.array([0.5, 0.5]))
        np.testing.assert_array_almost_equal(right, np.array([0.03333333, 0.15]))

        d = mix.mixalg()
        np.testing.assert_array_almost_equal(
            d["mix_den"], np.array([0.0, 0.0, 0.0, 0.0])
        )
        np.testing.assert_array_almost_equal(d["gradient"], np.array([0.0]))
        np.testing.assert_array_almost_equal(d["p"], np.array([1.0]))
        np.testing.assert_array_almost_equal(d["grid"], np.array([11.27659574]))

        assert d["k"] == 1
        assert d["accuracy"] == 1.0

        left, right = mix.getRateEstimates()
        np.testing.assert_array_almost_equal(
            left, np.array([0.0911574, 0.0911574, 0.0911574, 0.0911574])
        )
        np.testing.assert_array_almost_equal(right, np.array([1, 1, 1, 1]))



"""
A-DBSCAN implementation
"""

__author__ = "Dani Arribas-Bel <daniel.arribas.bel@gmail.com>"

import warnings
from collections import Counter

import numpy as np
import pandas
from libpysal.cg.alpha_shapes import alpha_shape_auto
from scipy.spatial import cKDTree
from sklearn.base import BaseEstimator as _BaseEstimator
from sklearn.base import ClusterMixin as _ClusterMixin
from sklearn.cluster import DBSCAN
from sklearn.neighbors import KNeighborsClassifier

__all__ = ["ADBSCAN", "remap_lbls", "ensemble", "get_cluster_boundary"]


class ADBSCAN(_ClusterMixin, _BaseEstimator):
    """
    A-DBSCAN, as introduced in :cite:`ab_gl_vm2020joue`.

    A-DSBCAN is an extension of the original DBSCAN algorithm that creates an
    ensemble of solutions generated by running DBSCAN on a random subset and
    "extending" the solution to the rest of the sample through
    nearest-neighbor regression.

    See the original reference (:cite:`ab_gl_vm2020joue`) for more details or
    the notebook guide for an illustration.
    ...

    Parameters
    ----------
    eps : float
        The maximum distance between two samples for them to be considered
        as in the same neighborhood.
    min_samples : int
        The number of samples (or total weight) in a neighborhood
        for a point to be considered as a core point. This includes the
        point itself.
    algorithm : {'auto', 'ball_tree', 'kd_tree', 'brute'}, optional
        The algorithm to be used by the NearestNeighbors module
        to compute pointwise distances and find nearest neighbors.
        See NearestNeighbors module documentation for details.
    n_jobs : int
        [Optional. Default=1] The number of parallel jobs to run. If -1, then
        the number of jobs is set to the number of CPU cores.
    pct_exact : float
        [Optional. Default=0.1] Proportion of the entire dataset
        used to calculate DBSCAN in each draw
    reps : int
        [Optional. Default=100] Number of random samples to draw in order to
        build final solution
    keep_solus : bool
        [Optional. Default=False] If True, the `solus` and `solus_relabelled`
        objects are kept, else it is deleted to save memory
    pct_thr : float
        [Optional. Default=0.9] Minimum proportion of replications that
        a non-noise label need to be assigned to an observation for that
        observation to be labelled as such

    Attributes
    ----------
    labels_ : array
        [Only available after `fit`] Cluster labels for each point in the
        dataset given to fit().
        Noisy (if the proportion of the most common label is < pct_thr)
        samples are given the label -1.
    votes : DataFrame
        [Only available after `fit`] Table indexed on `X.index` with
        `labels_` under the `lbls` column, and the frequency across draws of
        that label under `pct`
    solus : DataFrame, shape = [n, reps]
        [Only available after `fit`] Each solution of labels for every draw
    solus_relabelled : DataFrame, shape = [n, reps]
        [Only available after `fit`] Each solution of labels for
        every draw, relabelled to be consistent across solutions

    Examples
    --------
    >>> import pandas
    >>> from esda.adbscan import ADBSCAN
    >>> import numpy as np
    >>> np.random.seed(10)
    >>> db = pandas.DataFrame({'X': np.random.random(25), \
                               'Y': np.random.random(25) \
                              })

    ADBSCAN can be run following scikit-learn like API as:

    >>> np.random.seed(10)
    >>> clusterer = ADBSCAN(0.03, 3, reps=10, keep_solus=True)
    >>> _ = clusterer.fit(db)
    >>> clusterer.labels_
    array(['-1', '-1', '-1', '0', '-1', '-1', '-1', '0', '-1', '-1', '-1',
           '-1', '-1', '-1', '0', '0', '0', '-1', '0', '-1', '0', '-1', '-1',
           '-1', '-1'], dtype=object)

    We can inspect the winning label for each observation, as well as the
    proportion of votes:

    >>> print(clusterer.votes.head().to_string())
      lbls  pct
    0   -1  0.7
    1   -1  0.5
    2   -1  0.7
    3    0  1.0
    4   -1  0.7

    If you have set the option to keep them, you can even inspect each
    solution that makes up the ensemble:

    >>> print(clusterer.solus.head().to_string())
      rep-00 rep-01 rep-02 rep-03 rep-04 rep-05 rep-06 rep-07 rep-08 rep-09
    0      0      1      1      0      1      0      0      0      1      0
    1      1      1      1      1      0      1      0      1      1      1
    2      0      1      1      0      0      1      0      0      1      0
    3      0      1      1      0      0      1      1      1      0      0
    4      0      1      1      1      0      1      0      1      0      1



    If we select only one replication and the proportion of the entire dataset
    that is sampled to 100%, we obtain a traditional DBSCAN:

    >>> clusterer = ADBSCAN(0.2, 5, reps=1, pct_exact=1)
    >>> np.random.seed(10)
    >>> _ = clusterer.fit(db)
    >>> clusterer.labels_
    array(['0', '-1', '0', '0', '0', '-1', '-1', '0', '-1', '-1', '0', '-1',
           '-1', '-1', '0', '0', '0', '-1', '0', '0', '0', '-1', '-1', '0',
           '-1'], dtype=object)

    """

    def __init__(
        self,
        eps,
        min_samples,
        algorithm="auto",
        n_jobs=1,
        pct_exact=0.1,
        reps=100,
        keep_solus=False,
        pct_thr=0.9,
    ):
        self.eps = eps
        self.min_samples = min_samples
        self.algorithm = algorithm
        self.reps = reps
        self.n_jobs = n_jobs
        self.pct_exact = pct_exact
        self.pct_thr = pct_thr
        self.keep_solus = keep_solus

    def fit(self, X, y=None, sample_weight=None, xy=["X", "Y"]):  # noqa: ARG002
        """
        Perform ADBSCAN clustering from fetaures
        ...

        Parameters
        ----------
        X               : DataFrame
                          Features
        sample_weight   : Series, shape (n_samples,)
                          [Optional. Default=None] Weight of each sample, such
                          that a sample with a weight of at least ``min_samples``
                          is by itself a core sample; a sample with negative
                          weight may inhibit its eps-neighbor from being core.
                          Note that weights are absolute, and default to 1.
        xy              : list
                          [Default=`['X', 'Y']`] Ordered pair of names for XY
                          coordinates in `xys`
        y               : Ignored
        """
        n = X.shape[0]
        zfiller = len(str(self.reps))
        solus = pandas.DataFrame(
            np.zeros((X.shape[0], self.reps), dtype=str),
            index=X.index,
            columns=[f"rep-{str(i).zfill(zfiller)}" for i in range(self.reps)],
        )
        # Multi-core implementation of parallel draws
        if (self.n_jobs == -1) or (self.n_jobs > 1):
            # Set different parallel seeds!!!
            warnings.warn(
                "Multi-core implementation only works on relabelling solutions. "
                "Execution of draws is still sequential.",
                UserWarning,
                stacklevel=2,
            )
            for i in range(self.reps):
                pars = (
                    n,
                    X,
                    sample_weight,
                    xy,
                    self.pct_exact,
                    self.eps,
                    self.min_samples,
                    self.algorithm,
                    self.n_jobs,
                )
                lbls_pred = _one_draw(pars)
                solus.iloc[:, i] = lbls_pred
        else:
            for i in range(self.reps):
                pars = (
                    n,
                    X,
                    sample_weight,
                    xy,
                    self.pct_exact,
                    self.eps,
                    self.min_samples,
                    self.algorithm,
                    self.n_jobs,
                )
                lbls_pred = _one_draw(pars)
                solus.iloc[:, i] = lbls_pred

        solus_relabelled = remap_lbls(solus, X, xy=xy, n_jobs=self.n_jobs)
        self.votes = ensemble(solus_relabelled)
        lbls = self.votes["lbls"].values
        lbl_type = type(solus.iloc[0, 0])
        lbls[self.votes["pct"] < self.pct_thr] = lbl_type(-1)
        self.labels_ = lbls
        if not self.keep_solus:
            del solus
            del solus_relabelled
        else:
            self.solus = solus
            self.solus_relabelled = solus_relabelled
        return self


def _one_draw(pars):
    n, X, sample_weight, xy, pct_exact, eps, min_samples, algorithm, n_jobs = pars
    rids = np.arange(n)
    np.random.shuffle(rids)
    rids = rids[: int(n * pct_exact)]

    X_thin = X.iloc[rids, :]

    thin_sample_weight = None
    if sample_weight is not None:
        thin_sample_weight = sample_weight.iloc[rids]

    min_samples = min_samples * pct_exact
    min_samples = 1 if min_samples < 1 else int(np.floor(min_samples))

    dbs = DBSCAN(
        eps=eps,
        min_samples=min_samples,
        algorithm=algorithm,
        n_jobs=n_jobs,
    ).fit(X_thin[xy], sample_weight=thin_sample_weight)
    lbls_thin = pandas.Series(dbs.labels_.astype(str), index=X_thin.index)

    NR = KNeighborsClassifier(n_neighbors=1)
    NR.fit(X_thin[xy], lbls_thin)
    lbls_pred = pandas.Series(NR.predict(X[xy]), index=X.index)
    return lbls_pred


def remap_lbls(solus, xys, xy=["X", "Y"], n_jobs=1):
    """
    Remap labels in solutions so they are comparable (same label
    for same cluster)
    ...

    Parameters
    ----------
    solus       : DataFrame
                  Table with labels for each point (row) and solution (column)
    xys         : DataFrame
                  Table including coordinates
    xy          : list
                  [Default=`['X', 'Y']`] Ordered pair of names for XY
                  coordinates in `xys`
    n_jobs      : int
                  [Optional. Default=1] The number of parallel jobs to run. If
                  -1, then the number of jobs is set to the number of CPU
                  cores.

    Returns
    -------
    onel_solus  : DataFrame
                  Table with original solutions remapped to consolidated
                  labels across all the columns

    Examples
    --------

    >>> import pandas
    >>> db = pandas.DataFrame({"X": [0, 0.1, 4, 6, 5], \
                               "Y": [0, 0.2, 5, 7, 5] \
                              })
    >>> solus = pandas.DataFrame({"rep-00": [0, 0, 7, 7, -1], \
                                  "rep-01": [4, 4, -1, 6, 6], \
                                  "rep-02": [5, 5, 8, 8, 8] \
                                 })
    >>> print(remap_lbls(solus, db).to_string())
       rep-00  rep-01  rep-02
    0       0       0       0
    1       0       0       0
    2       7      -1       7
    3       7       7       7
    4      -1       7       7
    """
    # N. of clusters by solution
    ns_clusters = solus.apply(lambda x: x.unique().shape[0])
    # Pick reference solution as one w/ max N. of clusters
    ref = ns_clusters[ns_clusters == ns_clusters.max()].iloc[[0]].index[0]
    lbl_type = type(solus[ref].iloc[0])
    # Obtain centroids of reference solution
    ref_centroids = (
        xys.groupby(solus[ref])[xy]
        .apply(lambda xys: xys.mean())
        .drop(lbl_type(-1), errors="ignore")
    )
    # Only continue if any solution
    if ref_centroids.shape[0] > 0:
        # Build KDTree and setup results holder
        ref_kdt = cKDTree(ref_centroids)
        remapped_solus = pandas.DataFrame(
            np.zeros(solus.shape, dtype=lbl_type),
            index=solus.index,
            columns=solus.columns,
        )
        if (n_jobs == -1) or (n_jobs > 1):
            pool = _setup_pool(n_jobs)
            s_ids = solus.drop(ref, axis=1).columns.tolist()
            to_loop_over = [(solus[s], ref_centroids, ref_kdt, xys, xy) for s in s_ids]
            remapped = pool.map(_remap_n_expand, to_loop_over)
            remapped_df = pandas.concat(remapped, axis=1)
            if (
                remapped_df.isna().any().any()
                and remapped_solus[s_ids[0]].dtype.kind == "i"
            ):
                remapped_solus = remapped_solus.astype(float)
            remapped_solus.loc[:, s_ids] = remapped_df
        else:
            for s in solus.drop(ref, axis=1):
                # -
                pars = (solus[s], ref_centroids, ref_kdt, xys, xy)
                remap_ids = _remap_lbls_single(pars)
                # -
                mapped_values = solus[s].map(remap_ids)
                if mapped_values.isna().any() and remapped_solus[s].dtype.kind == "i":
                    remapped_solus[s] = remapped_solus[s].astype("float64")
                remapped_solus.loc[:, s] = mapped_values
        remapped_solus.loc[:, ref] = solus.loc[:, ref]
        return remapped_solus.fillna(lbl_type(-1)).astype(lbl_type)
    else:
        warnings.warn("No clusters identified.", UserWarning, stacklevel=2)
        return solus


def _remap_n_expand(pars):
    solus_s, ref_centroids, ref_kdt, xys, xy = pars
    remap_ids = _remap_lbls_single(pars)
    expanded = solus_s.map(remap_ids)
    return expanded


def _remap_lbls_single(pars):
    new_lbls, ref_centroids, ref_kdt, xys, xy = pars
    lbl_type = type(ref_centroids.index[0])
    # Cross-walk to cluster IDs
    ref_centroids_ids = pandas.Series(ref_centroids.index.values)
    # Centroids for new solution
    solu_centroids = (
        xys.groupby(new_lbls)[xy]
        .apply(lambda xys: xys.mean())
        .drop(lbl_type(-1), errors="ignore")
    )
    # Remapping from old to new labels
    _, nrst_ref_cl = ref_kdt.query(solu_centroids.values)
    remap_ids = pandas.Series(nrst_ref_cl, index=solu_centroids.index).map(
        ref_centroids_ids
    )
    return remap_ids


def ensemble(solus_relabelled):
    """
    Generate unique class prediction based on majority/hard voting
    ...

    Parameters
    ----------
    solus_relabelled  : DataFrame
                        Table with labels for each point (row) and solution
                        (column). Labels are assumed to be consistent across
                        solutions.

    Returns
    -------
    pred              : DataFrame
                        Table with one row per observation, a `lbls` column with the
                        winning label, and a `pct` column with the proportion of
                        times the winning label was voted

    Examples
    --------

    >>> import pandas
    >>> db = pandas.DataFrame({"X": [0, 0.1, 4, 6, 5], \
                               "Y": [0, 0.2, 5, 7, 5] \
                              })
    >>> solus = pandas.DataFrame({"rep-00": [0, 0, 7, 7, -1], \
                                  "rep-01": [4, 4, -1, 6, 6], \
                                  "rep-02": [5, 5, 8, 8, 8] \
                                 })
    >>> solus_rl = remap_lbls(solus, db)
    >>> print(round(ensemble(solus_rl), 2).to_string())
       lbls   pct
    0     0  1.00
    1     0  1.00
    2     7  0.67
    3     7  1.00
    4     7  0.67

    """

    counts = np.array(
        list(  # noqa: C417
            map(lambda a: Counter(a).most_common(1)[0], solus_relabelled.values)
        )
    )
    winner = counts[:, 0]
    votes = counts[:, 1].astype(int) / solus_relabelled.shape[1]
    pred = pandas.DataFrame(
        {"lbls": winner, "pct": votes}, index=solus_relabelled.index
    )
    return pred


def _setup_pool(n_jobs):
    """
    Set pool for multiprocessing
    ...

    Parameters
    ----------
    n_jobs      : int
                  The number of parallel jobs to run. If -1, then the number
                  of jobs is set to the number of CPU cores.
    """
    import multiprocessing as mp

    return mp.Pool(mp.cpu_count()) if n_jobs == -1 else mp.Pool(n_jobs)


def get_cluster_boundary(labels, xys, xy=["X", "Y"], n_jobs=1, crs=None, step=1):
    """
    Turn a set of labels associated with 2-D points into polygon boundaries
    for each cluster using the auto alpha shape algorithm
    (`libpysal.cg.alpha_shapes.alpha_shape_auto`)
    ...

    Parameters
    ----------
    labels      : Series
                  Cluster labels for each point in the dataset (noise
                  samples expressed as -1), indexed as `xys`
    xys         : DataFrame
                  Table including coordinates
    xy          : list
                  [Default=`['X', 'Y']`] Ordered pair of names for XY
                  coordinates in `xys`
    n_jobs      : int
                  [Optional. Default=1] The number of parallel jobs to run
                  for remapping. If -1, then the number of jobs is set to
                  the number of CPU cores.
    crs         : str
                  [Optional] Coordinate system
    step        : int
                  [Optional. Default=1]
                  Number of points in `xys` to jump ahead in the alpha
                  shape stage after checking whether the largest possible
                  alpha that includes the point and all the other ones
                  with smaller radii

    Returns
    -------
    polys       : GeoSeries
                  GeoSeries with polygons for each cluster boundary, indexed
                  on the cluster label

    Examples
    --------
    >>> import pandas
    >>> from esda.adbscan import ADBSCAN, get_cluster_boundary
    >>> import numpy as np
    >>> np.random.seed(10)
    >>> db = pandas.DataFrame({'X': np.random.random(25), \
                               'Y': np.random.random(25) \
                              })

    ADBSCAN can be run following scikit-learn like API as:

    >>> np.random.seed(10)
    >>> clusterer = ADBSCAN(0.03, 3, reps=10, keep_solus=True)
    >>> _ = clusterer.fit(db)
    >>> labels = pandas.Series(clusterer.labels_, index=db.index)
    >>> polys = get_cluster_boundary(labels, db)
    >>> polys[0].wkt
    'POLYGON ((0.7217553174317995 0.8192869956700687, 0.7605307121989587 0.9086488808086682, 0.9177741225129434 0.8568503024577332, 0.8126209616521135 0.6262871483113925, 0.6125260668293881 0.5475861559192435, 0.5425443680112613 0.7546476915298572, 0.7217553174317995 0.8192869956700687))'
    """  # noqa: E501

    try:
        from geopandas import GeoSeries
    except ModuleNotFoundError:

        def GeoSeries(data, index=None, crs=None):  # noqa: ARG001, N802
            return list(data)

    lbl_type = type(labels.iloc[0])
    noise = lbl_type(-1)
    ids_in_cluster = labels[labels != noise].index
    g = xys.loc[ids_in_cluster, xy].groupby(labels[ids_in_cluster])
    chunked_pts_step = []
    cluster_lbls = []
    for sub in g.groups:
        chunked_pts_step.append((xys.loc[g.groups[sub], xy].values, step))
        cluster_lbls.append(sub)
    if n_jobs == 1:
        polys = map(_asa, chunked_pts_step)
    else:
        pool = _setup_pool(n_jobs)
        polys = pool.map(_asa, chunked_pts_step)
        pool.close()
    polys = GeoSeries(polys, index=cluster_lbls, crs=crs)
    return polys


def _asa(pts_s):
    return alpha_shape_auto(pts_s[0], step=pts_s[1])



"""
Spatial autocorrelation for binary attributes

"""

__author__ = "Sergio J. Rey <srey@asu.edu> , Luc Anselin <luc.anselin@asu.edu>"

import warnings

import numpy as np
import pandas as pd
from libpysal.weights import W
from scipy.stats import chi2, chi2_contingency

from .crand import njit as _njit
from .tabular import _univariate_handler

__all__ = ["Join_Counts"]

PERMUTATIONS = 999


class Join_Counts:  # noqa: N801
    """Binary Join Counts


    Parameters
    ----------

    y               : array
                      binary variable measured across n spatial units
    w               : W | Graph
                      spatial weights instance as W or Graph aligned with y
    permutations    : int
                      number of random permutations for calculation of pseudo-p_values

    Attributes
    ----------
    y            : array
                   original variable
    w            : W
                   original w object
    permutations : int
                   number of permutations
    bb           : float
                   number of black-black joins
    ww           : float
                   number of white-white joins
    bw           : float
                   number of black-white joins
    J            : float
                   number of joins
    sim_bb       : array
                   (if permutations>0)
                   vector of bb values for permuted samples
    p_sim_bb     : array
                  (if permutations>0)
                   p-value based on permutations (one-sided)
                   null: spatial randomness
                   alternative: the observed bb is greater than under randomness
    mean_bb      : float
                   average of permuted bb values
    min_bb       : float
                   minimum of permuted bb values
    max_bb       : float
                   maximum of permuted bb values
    sim_bw       : array
                   (if permutations>0)
                   vector of bw values for permuted samples
    p_sim_bw     : array
                   (if permutations>0)
                   p-value based on permutations (one-sided)
                   null: spatial randomness
                   alternative: the observed bw is greater than under randomness
    mean_bw      : float
                   average of permuted bw values
    min_bw       : float
                   minimum of permuted bw values
    max_bw       : float
                   maximum of permuted bw values
    chi2         : float
                   Chi-square statistic on contingency table for join counts
    chi2_p       : float
                   Analytical p-value for chi2
    chi2_dof     : int
                   Degrees of freedom for analytical chi2
    crosstab     : DataFrame
                   Contingency table for observed join counts
    expected     : DataFrame
                   Expected contingency table for the null
    p_sim_chi2   : float
                   p-value for chi2 under random spatial permutations
    drop_islands : bool (default True)
        Whether or not to preserve islands as entries in the adjacency
        list. By default, observations with no neighbors do not appear
        in the adjacency list. If islands are kept, they are coded as
        self-neighbors with zero weight. See ``libpysal.weights.to_adjlist()``.



    Examples
    --------

    >>> import numpy as np
    >>> import libpysal
    >>> w = libpysal.weights.lat2W(4, 4)
    >>> y = np.ones(16)
    >>> y[0:8] = 0
    >>> np.random.seed(12345)
    >>> from esda.join_counts import Join_Counts
    >>> jc = Join_Counts(y, w)
    >>> jc.bb
    10.0
    >>> jc.bw
    4.0
    >>> jc.ww
    10.0
    >>> jc.J
    24.0
    >>> len(jc.sim_bb)
    999
    >>> round(jc.p_sim_bb, 3)
    0.003
    >>> round(np.mean(jc.sim_bb), 3)
    5.547
    >>> np.max(jc.sim_bb)
    10.0
    >>> np.min(jc.sim_bb)
    0.0
    >>> len(jc.sim_bw)
    999
    >>> jc.p_sim_bw
    1.0
    >>> np.mean(jc.sim_bw)
    12.811811811811811
    >>> np.max(jc.sim_bw)
    24.0
    >>> np.min(jc.sim_bw)
    7.0
    >>> round(jc.chi2_p, 3)
    0.004
    >>> jc.p_sim_chi2
    0.002

    Notes
    -----
    Technical details and derivations can be found in :cite:`cliff81`.

    """

    def __init__(self, y, w, permutations=PERMUTATIONS, drop_islands=True):
        y = np.asarray(y).flatten()

        if isinstance(w, W):
            w.transform = "b"  # ensure we have binary weights
            self.w = w
            self.adj_list = self.w.to_adjlist(
                remove_symmetric=False, drop_islands=drop_islands
            )
        else:
            w = w.transform("b")
            self.w = w
            if drop_islands:
                self.adj_list = w.adjacency.drop(w.isolates).reset_index()
            else:
                self.adj_list = w.adjacency.reset_index()
        self.y = y
        self.permutations = permutations
        s0 = w.s0 if isinstance(w, W) else w._adjacency.sum()
        self.J = s0 / 2.0
        results = self.__calc(self.y)
        self.bb = results[0]
        self.ww = results[1]
        self.bw = results[2]
        self.chi2 = results[3]
        self.chi2_p = results[4]
        self.chi2_dof = results[5]
        self.autocorr_pos = self.bb + self.ww
        self.autocorr_neg = self.bw

        crosstab = pd.DataFrame(data=results[-1])
        id_names = ["W", "B"]
        idx = pd.Index(id_names, name="Focal")
        crosstab.set_index(idx, inplace=True)
        crosstab.columns = pd.Index(id_names, name="Neighbor")
        self.crosstab = crosstab
        expected = pd.DataFrame(data=results[6])
        expected.set_index(idx, inplace=True)
        expected.columns = pd.Index(id_names, name="Neighbor")
        self.expected = expected
        self.calc = self.__calc

        if permutations:
            sim = []
            i = 0
            while i < permutations:
                try:
                    res = self.__calc(np.random.permutation(self.y))
                    sim.append(res)
                    i += 1
                except ValueError:
                    # expected count of 0 -> inadmissible
                    pass
            sim_jc = np.array(sim, dtype=object)
            self.sim_bb = sim_jc[:, 0]
            self.min_bb = np.min(self.sim_bb)
            self.mean_bb = np.mean(self.sim_bb)
            self.max_bb = np.max(self.sim_bb)
            self.sim_bw = sim_jc[:, 2]
            self.min_bw = np.min(self.sim_bw)
            self.mean_bw = np.mean(self.sim_bw)
            self.max_bw = np.max(self.sim_bw)
            self.sim_autocurr_pos = sim_jc[:, 0] + sim_jc[:, 1]
            self.sim_autocurr_neg = sim_jc[:, 2]
            self.sim_chi2 = sim_jc[:, 3]

            stat = (self.autocorr_pos - np.mean(self.sim_autocurr_pos)) ** 2 / np.mean(
                self.sim_autocurr_pos
            ) ** 2 + (
                self.autocorr_neg - np.mean(self.sim_autocurr_neg)
            ) ** 2 / np.mean(self.sim_autocurr_pos) ** 2
            self.sim_autocorr_chi2 = 1 - chi2.cdf(stat, 1)

            p_sim_bb = self.__pseudop(self.sim_bb, self.bb)
            p_sim_bw = self.__pseudop(self.sim_bw, self.bw)
            p_sim_chi2 = self.__pseudop(self.sim_chi2, self.chi2)
            p_sim_autocorr_pos = self.__pseudop(
                self.sim_autocurr_pos, self.autocorr_pos
            )
            p_sim_autocorr_neg = self.__pseudop(
                self.sim_autocurr_neg, self.autocorr_neg
            )
            self.p_sim_bb = p_sim_bb
            self.p_sim_bw = p_sim_bw
            self.p_sim_chi2 = p_sim_chi2
            self.p_sim_autocorr_pos = p_sim_autocorr_pos
            self.p_sim_autocorr_neg = p_sim_autocorr_neg

    def __calc(self, z):
        adj_list = self.adj_list
        zseries = pd.Series(
            z,
            index=self.w.id_order if hasattr(self.w, "id_order") else self.w.unique_ids,
        )
        focal = zseries.loc[adj_list.focal].values
        neighbor = zseries.loc[adj_list.neighbor].values
        sim = focal == neighbor
        dif = 1 - sim
        bb = (focal * sim).sum() / 2
        ww = ((1 - focal) * sim).sum() / 2
        bw = (focal * dif).sum() / 2
        wb = ((1 - focal) * dif).sum() / 2
        table = [[ww, wb], [bw, bb]]
        chi2 = chi2_contingency(table)
        stat, pvalue, dof, expected = chi2
        return (bb, ww, bw + wb, stat, pvalue, dof, expected, np.array(table))

    def __pseudop(self, sim, jc):
        above = sim >= jc
        larger = sum(above)
        psim = (larger + 1.0) / (self.permutations + 1.0)
        return psim

    @property
    def _statistic(self):
        return self.bw

    @classmethod
    def by_col(
        cls, df, cols, w=None, inplace=False, pvalue="sim", outvals=None, **stat_kws
    ):
        """
        Function to compute a Join_Count statistic on a dataframe

        Parameters
        ----------
        df : pandas.DataFrame
            a pandas dataframe with a geometry column
        cols : string or list of string
            name or list of names of columns to use to compute the statistic
        w : W | Graph
            spatial weights instance as W or Graph aligned with the dataframe. If not
            provided, this is searched for in the dataframe's metadata
        inplace : bool
            a boolean denoting whether to operate on the dataframe inplace or to
            return a series contaning the results of the computation. If
            operating inplace, the derived columns will be named
            'column_join_count'
        pvalue : string
            a string denoting which pvalue should be returned. Refer to the
            the Join_Count statistic's documentation for available p-values
        outvals : list of strings
            list of arbitrary attributes to return as columns from the
            Join_Count statistic
        **stat_k : dict
            options to pass to the underlying statistic. For this, see the
            documentation for the Join_Count statistic.

        Returns
        --------
        If inplace, None, and operation is conducted on
        dataframe in memory. Otherwise, returns a copy of the
        dataframe with the relevant columns attached.

        """

        msg = (
            "The `.by_col()` methods are deprecated and will be "
            "removed in a future version of `esda`."
        )
        warnings.warn(msg, FutureWarning, stacklevel=2)

        if outvals is None:
            outvals = []
            outvals.extend(["bb", "p_sim_bw", "p_sim_bb"])
            pvalue = ""
        return _univariate_handler(
            df,
            cols,
            w=w,
            inplace=inplace,
            pvalue=pvalue,
            outvals=outvals,
            stat=cls,
            swapname="bw",
            **stat_kws,
        )


# --------------------------------------------------------------
# Conditional Randomization Function Implementations
# --------------------------------------------------------------
@_njit(fastmath=True)
def _local_join_count_crand():
    raise NotImplementedError



"""Geary Unittest."""

import numpy as np
import pytest
from libpysal import examples, graph
from libpysal.io import open as popen
from libpysal.weights import Rook

from .. import geary

w = Rook.from_shapefile(examples.get_path("columbus.shp"))
w.transform = "r"

parametrize_w = pytest.mark.parametrize(
    "w",
    [
        w,
        graph.Graph.from_W(w),
    ],
    ids=["W", "Graph"],
)


class TestGeary:
    """Geary class for unit tests."""

    def setup_method(self):
        f = popen(examples.get_path("columbus.dbf"))
        self.y = np.array(f.by_col["CRIME"])

    @parametrize_w
    def test_default(self, w):
        c = geary.Geary(self.y, w, permutations=0)
        np.testing.assert_allclose(c.C, 0.5154408058652411)
        np.testing.assert_allclose(c.EC, 1.0)
        np.testing.assert_allclose(c.VC_norm, 0.011403109626468939)
        np.testing.assert_allclose(c.seC_norm, 0.10678534368755357)
        np.testing.assert_allclose(c.p_norm, 2.8436375936967054e-06)
        np.testing.assert_allclose(c.z_norm, -4.5376938201608)

        np.testing.assert_allclose(c.VC_rand, 0.010848882767131886)
        np.testing.assert_allclose(c.p_rand, 1.6424069196305004e-06)
        np.testing.assert_allclose(c.z_rand, -4.652156651668989)

        np.random.seed(12345)
        c = geary.Geary(self.y, w, permutations=999)
        np.testing.assert_allclose(c.C, 0.5154408058652411)
        np.testing.assert_allclose(c.EC, 1.0)
        np.testing.assert_allclose(c.VC_norm, 0.011403109626468939)
        np.testing.assert_allclose(c.seC_norm, 0.10678534368755357)
        np.testing.assert_allclose(c.p_norm, 2.8436375936967054e-06)
        np.testing.assert_allclose(c.z_norm, -4.5376938201608)

        np.testing.assert_allclose(c.VC_rand, 0.010848882767131886)
        np.testing.assert_allclose(c.p_rand, 1.6424069196305004e-06)
        np.testing.assert_allclose(c.z_rand, -4.652156651668989)

        np.testing.assert_allclose(c.EC_sim, 0.9981883958193233)
        np.testing.assert_allclose(c.VC_sim, 0.010631247074115058)
        np.testing.assert_allclose(c.p_sim, 0.001)
        np.testing.assert_allclose(c.p_z_sim, 1.4207015378575605e-06)

    @parametrize_w
    def test_by_col(self, w):
        import pandas as pd

        df = pd.DataFrame(self.y, columns=["y"])
        np.random.seed(12345)
        r1 = geary.Geary.by_col(df, ["y"], w=w, permutations=999)
        this_geary = np.unique(r1.y_geary.values)
        this_pval = np.unique(r1.y_p_sim.values)
        c = geary.Geary(self.y, w, permutations=999)
        np.testing.assert_allclose(this_geary, c.C)
        np.testing.assert_allclose(this_pval, c.p_sim)



import contextlib
import warnings

import geopandas
import numpy
import pandas
import shapely
from packaging.version import Version

from .crand import njit, prange

__author__ = (
    "Martin Fleischmann <martin@fleischmann.net>",
    "Levi John Wolf <levi.john.wolf@gmail.com>",
    "Alan Murray <amurray@ucsb.edu>",
    "Jiwan Baik <jiwon.baik@geog.ucsb.edu>",
)


# -------------------- UTILITIES --------------------#
# shapely.orient_polygons not available in shapely<2.1. Check version, and if 
# not available, use shapely.normalize (to make exterior rings clockwise) + 
# shapely.reverse. This is about 30% slower than shapely.orient_polygons.
# Remove and use shapely.orient_polygons when shapely 2.1+ is required.

if Version(shapely.__version__) >= Version("2.1.0"):
    orient_polygons = shapely.orient_polygons
else:
    def orient_polygons(geometry, exterior_cw=False):
        g = shapely.normalize(geometry)
        if not exterior_cw:
            g = shapely.reverse(g)
        return g
    
def _cast(collection):
    """
    Cast a collection to a shapely geometry array.
    """

    if isinstance(collection, geopandas.GeoSeries | geopandas.GeoDataFrame):
        return numpy.asarray(collection.geometry.array)
    elif isinstance(collection, numpy.ndarray | list):
        return numpy.asarray(collection)
    else:
        return numpy.array([collection])

def _cast_pts_as_array(x):
    """
    Accepts:
      - array-like of shape (2,) or (n, 2) of real-valued numerics
      - shapely.geometry.Point or iterable of Points
      - geopandas.GeoSeries of Points

    Returns:
      - ndarray of shape (2,) or (n, 2), dtype float

    Elements must be real-valued numerics (no complex).
    Geometry inputs must be Points only; non-Point geometries are rejected.
    """

    # Handle GeoSeries
    if isinstance(x, geopandas.GeoSeries):
        if not all(x.geom_type == "Point"):
            raise TypeError("All geometries in GeoSeries must be Points")
        coords = shapely.get_coordinates(x)
        if coords.shape == (1, 2):
            coords = coords[0]
        return coords

    # Handle Single Point
    if isinstance(x, shapely.Point):
        return numpy.array([x.x, x.y], dtype=float)

    # Handle Iterable of Points
    if hasattr(x, "__iter__") and all(isinstance(p, shapely.Point) for p in x):
        coords = shapely.get_coordinates(x)
        return coords

    # Handle Array-Like
    try:
        arr = numpy.asarray(x, dtype=float)
    except Exception as e:
        raise TypeError("Input must be array-like") from e

    # Shape validation
    if arr.ndim == 1:
        if arr.shape != (2,):
            raise ValueError(f"Expected shape (2,), got {arr.shape}")
    elif arr.ndim == 2:
        if arr.shape[1] != 2:
            raise ValueError(f"Expected shape (n, 2), got {arr.shape}")
    else:
        raise ValueError(f"Expected 1D or 2D input, got {arr.ndim}D")

    # Type validation: real numbers only
    if not numpy.issubdtype(arr.dtype, numpy.floating) and not numpy.issubdtype(arr.dtype, numpy.integer):
        raise TypeError("Elements must be floats or ints")

    return arr


def get_angles(collection, return_indices=False):
    """
    Get the angles pertaining to each vertex of a set of polygons.
    This assumes the input are polygons.

    Parameters
    ----------
    ga  :   shapely geometry array
        array of polygons/multipolygons
    return_indices  :   bool (Default: False)
        whether to return the indices relating each geometry to a polygon

    Returns
    -------
    angles between triples of points on each geometry, as well as the indices
    relating angles to input geometries (if requested).

    See the Notes for information on the shape of angles and indices.

    Notes
    -------
    If a geometry has n coordinates and k parts, the array will be n - k.
    If each geometry has n_i coordinates, then let N be a vector storing
    those counts (computed, for example, using shapely.get_num_coordinates(ga)).
    Likewise, let K be a vector storing the number of parts each geometry has, k_i
    (computed, for example, using shapely.get_num_geometries(ga))

    Then, the output is of shape (N - K).sum()

    """
    ga = _cast(collection)
    exploded = shapely.get_parts(ga)
    coords = shapely.get_coordinates(exploded)
    n_coords_per_geom = shapely.get_num_coordinates(exploded)
    angles = numpy.asarray(_get_angles(coords, n_coords_per_geom))
    if return_indices:
        return angles, numpy.repeat(
            numpy.arange(len(ga)),
            shapely.get_num_coordinates(ga) - shapely.get_num_geometries(ga),
        )
    else:
        return angles


@njit
def _get_angles(points, n_coords_per_geom):
    """
    Iterate over points in a set of geometries.
    This assumes that the input geometries are simple, not multi!
    """
    # Start at the first point of the first geometry
    offset = 0
    on_geom = 0
    on_coord = 0
    result = []
    while True:
        # if we're on the last point before the closure point,
        if on_coord == (n_coords_per_geom[on_geom] - 1):
            # set the offset to start on the first point of the next geometry
            offset += on_coord + 1
            on_geom += 1
            on_coord = 0
            # if we're now done with all geometries, exit
            if on_geom == len(n_coords_per_geom):
                break
            else:
                # and continue to the next iteration.
                continue
        # construct the triple so that we wrap around and avoid the closure point
        left_ix = offset + on_coord % (n_coords_per_geom[on_geom] - 1)
        center_ix = offset + (on_coord + 1) % (n_coords_per_geom[on_geom] - 1)
        right_ix = offset + (on_coord + 2) % (n_coords_per_geom[on_geom] - 1)
        # grab the actual coordinates corresponding to the triple
        left = points[left_ix]
        center = points[center_ix]
        right = points[right_ix]
        # build the line segments originating at center
        a = left - center
        b = right - center
        # compute the angle between the segments
        angle = numpy.arctan2(a[0] * b[1] - a[1] * b[0], numpy.dot(a, b))
        result.append(angle)
        on_coord += 1
    return result


# -------------------- IDEAL SHAPE MEASURES -------------------- #

def isoperimetric_quotient(collection):
    """
    The Isoperimetric quotient, defined as the ratio of a polygon's area to the
    area of the equi-perimeter circle.

    Parameters
    ----------
    collection : GeoSeries, GeoDataFrame, numpy.ndarray, list
        Input collection of polygons.

    Returns
    -------

    numpy.ndarray
        An array of the same length as the input collection, containing the 
        Isoperimetric quotient for each polygon in the collection.

    Notes
    -----

    Altman's :math:`PA_1` measure :cite:`altman1998Districting`.

    The formula is given by:

    .. math::
        IPQ = \\frac{4 \\pi A}{P^2}

    Where :math:`A` is the area of the polygon and :math:`P` is the perimeter of the polygon.

    The :math:`IPQ` is scale invariant and due to the inclusion of :math:`\\pi` in the formula, 
    it is bounded between 0 and 1, with 1 representing a perfect circle, the most compact shape
    by this measure.
    
    """

    ga = _cast(collection)
    return (4 * numpy.pi * shapely.area(ga)) / (shapely.measurement.length(ga) ** 2)


def isoareal_quotient(collection):
    """
    The Isoareal quotient, defined as the ratio of a polygon's perimeter to the
    perimeter of the equi-areal circle.

    Parameters
    ----------
    collection : GeoSeries, GeoDataFrame, numpy.ndarray, list
        Input collection of polygons.

    Returns
    -------

    numpy.ndarray
        An array of the same length as the input collection, containing the 
        Isoareal quotient for each polygon in the collection.
        
    Notes
    -----

    Altman's :math:`PA_3` measure :cite:`altman1998Districting`.

    The formula is given by:

    .. math::
        IAQ = \\frac{2 \\sqrt{\\pi A}}{P}

    Where :math:`A` is the area of the polygon and :math:`P` is the perimeter of the polygon.

    With some manipulation, :math:`IAQ` can also be expressed as the square root of the Isoperimetric quotient, given by

    .. math::
        IAQ = \\frac{2 \\sqrt{\\pi A}}{P}
            = \\sqrt{\\frac{(2 \\sqrt{\\pi A})^2}{P^2}}
            = \\sqrt{\\frac{4 \\pi A}{P^2}}
            = \\sqrt{IPQ}

    Therefore, `isoareal_quotient` is implemented as `numpy.sqrt(isoperimetric_quotient(collection))`. 
    Importantly, this means that the :math:`IAQ` and :math:`IPQ` will rank shapes identically.

    The :math:`IAQ` is scale invariant and due to the inclusion of :math:`\\pi` in the formula, 
    it is bounded between 0 and 1, with 1 representing a perfect circle, the most compact shape
    by this measure.
    
    """
    return numpy.sqrt(isoperimetric_quotient(collection))


def minimum_bounding_circle_ratio(collection):
    """
    The Reock compactness measure, defined by the ratio of areas between the
    minimum bounding/containing circle of a shape and the shape itself.

    Measure A1 in :cite:`altman1998Districting`,
    cited for Frolov (1974), but earlier from Reock
    (1963)
    """
    ga = _cast(collection)
    mbca = (shapely.minimum_bounding_radius(ga) ** 2) * numpy.pi
    return shapely.area(ga) / mbca


def radii_ratio(collection):
    """
    The Flaherty & Crumplin (1992) index, OS_3 in :cite:`altman1998Districting`.

    The ratio of the radius of the equi-areal circle to the radius of the MBC
    """
    ga = _cast(collection)
    r_eac = numpy.sqrt(shapely.area(ga) / numpy.pi)
    r_mbc = shapely.minimum_bounding_radius(ga)
    return r_eac / r_mbc


def diameter_ratio(collection, rotated=True):
    """
    The Flaherty & Crumplin (1992) length-width measure, stated as measure LW_7
    in :cite:`altman1998Districting`.

    It is given as the ratio between the minimum and maximum shape diameter.
    """
    ga = _cast(collection)
    if rotated:
        box = shapely.minimum_rotated_rectangle(ga)
        coords = shapely.get_coordinates(box)
        a, b, _, d = (coords[0::5], coords[1::5], coords[2::5], coords[3::5])
        widths = numpy.sqrt(numpy.sum((a - b) ** 2, axis=1))
        heights = numpy.sqrt(numpy.sum((a - d) ** 2, axis=1))
    else:
        box = shapely.bounds(ga)
        (xmin, xmax), (ymin, ymax) = box[:, [0, 2]].T, box[:, [1, 3]].T
        widths, heights = numpy.abs(xmax - xmin), numpy.abs(ymax - ymin)
    return numpy.minimum(widths, heights) / numpy.maximum(widths, heights)


def length_width_diff(collection):
    """
    The Eig & Seitzinger (1981) shape measure, defined as:

    L - W

    Where L is the maximal North-South extent and W is the maximal East-West
    extent.

    Defined as measure LW_5 in :cite:`altman1998Districting`
    """
    ga = _cast(collection)
    box = shapely.bounds(ga)
    (xmin, xmax), (ymin, ymax) = box[:, [0, 2]].T, box[:, [1, 3]].T
    width, height = numpy.abs(xmax - xmin), numpy.abs(ymax - ymin)
    return height - width


def boundary_amplitude(collection):
    """
    The boundary amplitude (adapted from Wang & Huang (2012)) is the
    length of the boundary of the convex hull divided by the length of the
    boundary of the original shape.

    Notes
    -----

    This is inverted from Wang & Huang (2012) in order to provide a value
    between zero and one, like many of the other ideal shape-based indices.
    """
    ga = _cast(collection)
    return shapely.length(shapely.convex_hull(ga)) / shapely.length(ga)


def convex_hull_ratio(collection):
    """
    ratio of the area of the convex hull to the area of the shape itself

    Altman's A_3 measure, from Neimi et al 1991.
    """
    ga = _cast(collection)
    return shapely.area(ga) / shapely.area(shapely.convex_hull(ga))


def fractal_dimension(collection, support="hex"):
    """
    The fractal dimension of the boundary of a shape, assuming a given
    spatial support for the geometries.

    Note that this derivation assumes a specific ideal spatial support
    for the polygon, and is thus may not return valid results for
    complex or highly irregular geometries.
    """
    ga = _cast(collection)
    P = shapely.length(ga)
    A = shapely.area(ga)
    if support == "hex":
        return 2 * numpy.log(P / 6) / numpy.log(A / (3 * numpy.sin(numpy.pi / 3)))
    elif support == "square":
        return 2 * numpy.log(P / 4) / numpy.log(A)
    elif support == "circle":
        return 2 * numpy.log(P / (2 * numpy.pi)) / numpy.log(A / numpy.pi)
    else:
        raise ValueError(
            "The support argument must be one of 'hex', 'circle', or 'square', "
            f"but {support} was provided."
        )


def squareness(collection):
    """
    Measures how different is a given shape from an equi-areal square

    The index is close to 0 for highly irregular shapes and to 1.3 for circular shapes.
    It equals 1 for squares.

    .. math::
        \\begin{equation}
        \\frac{
            \\sqrt{A}}{P^{2}}
            \\times
            \\frac{\\left(4 \\sqrt{\\left.A\\right)}^{2}\\right.}{\\sqrt{A}}
            =
            \\frac{\\left(4 \\sqrt{A}\\right)^{2}}{P{ }^{2}}
            =
            \\left(\\frac{4 \\sqrt{A}}{P}\\right)^{2}
        \\end{equation}

    where :math:`A` is the area and :math:`P` is the perimeter.

    Notes
    -----
    Implementation follows :cite:`basaraner2017`.

    """
    ga = _cast(collection)
    return ((numpy.sqrt(shapely.area(ga)) * 4) / shapely.length(ga)) ** 2


def rectangularity(collection):
    """
    Ratio of the area of the shape to the area
    of its minimum bounding rotated rectangle

    Reveals a polygon’s degree of being curved inward.

    .. math::
        \\frac{A}{A_{MBR}}

    where :math:`A` is the area and :math:`A_{MBR}`
    is the area of minimum bounding
    rotated rectangle.

    Notes
    -----
    Implementation follows :cite:`basaraner2017`.
    """
    ga = _cast(collection)
    return shapely.area(ga) / shapely.area(shapely.minimum_rotated_rectangle(ga))


def shape_index(collection):
    """
    Schumm’s shape index (Schumm (1956) in MacEachren 1985)

    .. math::
        {\\sqrt{{A} \\over {\\pi}}} \\over {R}

    where :math:`A` is the area and :math:`R` is the radius of the minimum bounding
    circle.

    Notes
    -----
    Implementation follows :cite:`maceachren1985compactness`.

    """
    ga = _cast(collection)
    return numpy.sqrt(shapely.area(ga) / numpy.pi) / shapely.minimum_bounding_radius(ga)


def equivalent_rectangular_index(collection):
    """
    Deviation of a polygon from an equivalent rectangle

    .. math::
        \\frac{\\sqrt{A}}{A_{MBR}}
        \\times
        \\frac{P_{MBR}}{P}

    where :math:`A` is the area, :math:`A_{MBR}` is the area of minimum bounding
    rotated rectangle, :math:`P` is the perimeter, :math:`P_{MBR}` is the perimeter
    of minimum bounding rotated rectangle.

    Notes
    -----
    Implementation follows :cite:`basaraner2017`.
    """
    ga = _cast(collection)
    box = shapely.minimum_rotated_rectangle(ga)
    return numpy.sqrt(shapely.area(ga) / shapely.area(box)) * (
        shapely.length(box) / shapely.length(ga)
    )


# -------------------- VOLMETRIC MEASURES ------------------- #


def form_factor(collection, height):
    """
    Computes volumetric compactness

    .. math::
        \\frac{A}{(A \\times H)^{\\frac{2}{3}}}

    where :math:`A` is the area and :math:`H` is polygon's
    height.

    Notes
    -----
    Implementation follows :cite:`bourdic2012`.
    """
    ga = _cast(collection)
    A = shapely.area(ga)
    V = A * height
    zeros = V == 0
    res = numpy.zeros(len(ga))
    res[~zeros] = A[~zeros] / (V[~zeros] ** (2 / 3))
    return res


# -------------------- INERTIAL MEASURES -------------------- #

def moment_of_inertia(collection, normalize=False, ref_pt=None):
    """
    Compute moment of inertia (second moment of area) per geometry.

    Parameters
    ----------
    collection : GeoSeries, GeoDataFrame, numpy.ndarray, list
        Input collection of polygons or multipolygons.
    normalize : bool, optional
        If True, returns moment normalized by reference cricle of same area. Default is False.
    ref_pt : GeoSeries, Shapely Point or list of Points, array-like of shape (2,) or (n, 2), optional
        If provided, shifts moment to be with respect to this point or points.
        The default behavior (default: ``None``) is to calculate the moment 
        about the centroid of each geometry. Points may be passed as
        as array-like of coordinates or point geometries. If a single point,
        all moments are calculated with respect to that reference point. If
        of length equal to collection, the moment for each geometry in 
        collection is calculated with respect to the reference point with
        the same index. To return moment about the origin, explicitly set to (0, 0).
        
    Returns
    -------
    numpy.ndarray
        Array of moment of inertia values for each geometry in the collection.

    Notes
    -----

    Calculates the moment of inertia of each geometry in the collection. Can handle
    polygons with holes and multipolygons by summing the moments of inertia of each part.
    The moment of inertia is calculated about the centroid of each geometry by default,
    but can also be calculated about a specified reference point or points.
        
    The moment of inertia is the variance of the distance of all points
    in a shape to a reference point, which can be interpreted as an axis of rotation 
    perpendicular to the plane of the shape. The first parameter should be a GeoDataFrame or 
    array-like of Polygons or MultiPolygons. If normalization is
    requested, the moment of inertia is compared with that of a circle of equal area. This
    is discussed in detail below.
    
    Moments of inertia with weights (such as population) may be calculated using 
    `moment_of_inertia_regions`. Without weights, the moment of inertia is known as the 
    area moment of inertia or the second moment of area, and can be quickly calculated from 
    the polygon vertices using the shoelace formula. The second moments of area about the 
    x and y axes are calculated as:
    
    .. math::
        I_x = \\frac{1}{12}\\sum_{i=0}^{n-1} (x_i y_{i+1} - x_{i+1}y_i)(y_i^2 + y_i y_{i+1} + y_{i+1}^2)
        
        I_y = \\frac{1}{12}\\sum_{i=0}^{n-1} (x_i y_{i+1} - x_{i+1}y_i)(x_i^2 + x_i x_{i+1} + x_{i+1}^2)
    
    where indices wrap around (n+1 = 1).

    The moments are then adjusted to be relative to a reference point using the 
    parallel axis theorem. The moment of inertia is then the sum of :math:`I_x` and 
    :math:`I_y`. In geographic contexts, the reference point is typically the 
    centroid, which is the default if no reference point is provided.
    However, other points of interest may be used, such as a capital city or the residence of
    a political representative. 

    The moment of inertia can be normalized to provide a compactness measure by comparing
    it with the moment of inertia of a circle of equal area. This is calculated
    as the ratio of the moment of inertia of the circle to that of the shape about its centroid:

    .. math::
        C_{MI} = \\frac{I_{circle}}{I_{shape}}
               = \\frac{A^2}{2 \\pi I_{shape}}

    where :math:`A` is the area of the polygon and :math:`I_{shape}` is the moment of inertia
    of the polygon. This formulation is from Li, et al (2013). The value of 
    :math:`C_{MI}` is bounded between 0 and 1, with 1 representing a perfect circle, 
    the most compact shape by this measure.
    
    References
    ----------
    .. [1] Godwin, A. N. 1980. "Simple Calculation of Moments of Inertia for Polygons."
        International Journal of Mathematical Education in Science and Technology 11 (4): 577–86. 
        https://doi.org/10.1080/0020739800110414.
    .. [2] Li, Wenwen, Michael F. Goodchild, and Richard Church. 2013. "An Efficient Measure 
        of Compactness for Two-Dimensional Shapes and Its Application in Regionalization 
        Problems." International Journal of Geographical Information Science 27 (6): 1227–50. 
        https://doi.org/10.1080/13658816.2012.752093.
    .. [2] https://en.wikipedia.org/wiki/Second_moment_of_area#List_of_second_moments_of_area

    """
    ga = _cast(collection)
    ga = orient_polygons(ga) # shapely.orient_polygons(ga) # 

    if ref_pt is not None:
        coords = _cast_pts_as_array(ref_pt)
        if not (coords.shape == (2,) or coords.shape == (len(ga), 2)):
            msg = f"`ref_pt` must be a single point (or coordinate pair) or one point (or coordinate pairs) per geometry in `collection` ({len(ga)})"
            raise ValueError(msg)            

    Js = []
    for i, geom in enumerate(ga):
        A, Cx, Cy, Ixx, Iyy, J = _moments_about_centroid([geom])
        if ref_pt is not None:
            if coords.shape == (2,):
                dx = Cx - coords[0]
                dy = Cy - coords[1]
            else: # Already tested, if not (2,), must be (len(ga), 2)
                dx = Cx - coords[i][0]
                dy = Cy - coords[i][1]
            J += A * (dx**2 + dy**2)
        if normalize:
            J = A**2 / (2 * numpy.pi * J)
        Js.append(J)

    return numpy.asarray(Js)

second_moment_of_area = moment_of_inertia  # Alias for users familiar with math/engineering terminology

def second_areal_moment(collection):
    # Alias to preserve old API.

    msg = "`second_areal_moment` is deprecated and will be removed in a "
    "future version. Use `moment_of_inertia` instead. The current function "
    "is an alias for `moment_of_inertia` which does not expose all parameters."
    
    warnings.warn(msg, DeprecationWarning, stacklevel=2)

    return moment_of_inertia(collection, normalize=False, ref_pt=None)

def moment_of_inertia_regions(collection, normalize=False, ref_pt=None, 
                              regions=None, weights=None):
    """
    Compute weighted moment of inertia per region. See Notes for behavior when
    either regions or weights are omitted.

    Parameters
    ----------
    collection : GeoSeries, GeoDataFrame, numpy.ndarray, list
        Input collection of polygons or multipolygons.
    normalize : bool, optional
        If True, returns moment normalized by reference cricle of same area and
        mass (sum of weights in the region). Default is False.
    ref_pt : GeoSeries, Shapely Point or list of Points, array-like of shape 
        (2,) or (n, 2), or dict of any of these, optional
        If provided, shifts moment to be with respect to this point or points.
        The default behavior (default: ``None``) is to calculate the moment 
        about the centroid of each region or geometry. If `regions` is
        provided, this must be a `dict` with one item per region, with the key
        equal to the region identifier and the value equal to a point geometry
        or point coordinates. See `moment_of_inertia` for details.
    regions : array-like, str, optional
        An iterable of region identifiers of the same length as `collection` 
        that each geometry is assigned to, or the name of a column in the 
        GeoDataFrame to use for region assignment. If None (default), moment of inertia 
        is calculated for each geometry in `collection` without regionalization.
    weights : array-like, optional
        An iterable of weights (e.g., population) of the same length as 
        `collection` that are applied to each geometry in `collection`, or the 
        name of a column in the GeoDataFrame to use for weights. If None (default),
        calculates second moment of area using the shoelace formula.

    Returns
    -------
    pandas.Series or numpy.ndarray
        If `regions` is provided, returns a `pandas.Series` indexed 
        by unique region IDs, containing moments per region. If `regions` is 
        omitted, retuns an `numpy.ndarray` of moments per geometry in `collection`.

    Calculates the mass moment of inertia for regions defined by assignment of
    geometries in the collection.
    
    Requires either a column name with region assignments or an array-like of
    region IDs. Weights can be provided as a column name or an array-like. If 
    either `regions` or `weights` is a string, `collection` must be a 
    GeoDataFrame and the column name must be a valid column in the GeoDataFrame. 
    If weights are not provided, geometries weighted by area, which is equivalent 
    to the second moment of area.

    Notes
    -----
    See `moment_of_inertia` for an introduction. Note that moment of 
    intertia and second moment of area are used interchangeably in some
    disciplines (see Li, et al 2013). In this discussion "mass moment of
    inertia" is used to represent a weighted moment of inertia, and "second
    moment of area" is used for an unweighted moment of inertia. This
    module's `moment_of_inertia` function returns the second moment of
    area only.

    This function extends the `moment_of_inertia` implementation to allow
    regionalization and/or weighting. Region identifiers are provided via
    the `regions` parameter. Weighting is provided via the `weights` parameter.

    If region identifiers are provided, the geometries in `collection` are 
    subareas within larger regions defined by the region identifier. The moment 
    of inertia (possibly weighted) is calculated for each region as a whole. If 
    omitted, moments are calculated for each geometry in `collection`.

    If weights are provided, the mass moment of inertia is calculated using a 
    value of interest, such as population, as the mass of the 
    shape. Implementation details vary with whether regions are provided
    or normalization is requested. This is discussed in detail below. First, 
    if weights are **not** provided, each geometry is weighted by area, and
    the result is equivalent to the second moment of area. If weights are
    provided and normalization is requested, the mass of the shape
    and the reference circle cancel, and the result is equivalent to 
    the normalized second moment of area.

    The mass moment of inertia of a point is equal to its mass times the 
    distance to a reference point squared. For a shape rotating about its
    centroid the moment of inertia is:

    .. math::
        I = \\sum_{i} m_i r_i^2

    where :math:`r_i` is the distance from the point to a reference point, 
    :math:`m_i` is the mass at each point, and there are an infinite number 
    of points filling the shape.

    The mass moment of inertia can be calculated for an area of uniform density 
    or an area of varying density. For areas of uniform density, this is the 
    equivalent of the second moment of area times the mass of the shape
    divided by the area. If `regions` is not provided, the mass moment of inertia
    is calculated for each geometry in `collection` as an area of uniform
    density. This is implemented as:


    .. math::
        I_M = I_A m / A

    where :math:`I_A` is the second moment of area (calculated by `moment_of_inertia`)
    and :math:`m` and :math:`A` are the mass and area of the shape, respectively.

    For a region of varying density, region identifiers must be provided via 
    the `regions` parameter. :math:`I_M` is calculated for each geometry in 
    `collection` per the equation above, with each geometry representing a 
    subarea of the region with mass given
    by `weights`. The mass moment of inertia of the region is then the sum of 
    the mass moments of inertia of each subarea shifted to the reference point
    (using the parallel axis theorem) by adding :math:`m d^2`, where :math:`d` 
    is the distance from the centroid of each subarea to the reference point 
    for the region.

    If reference points are not provided, the mass moment of inertia is 
    calculated with respect to the centroid of each region (calculated by the
    algorithm). If provided, `ref_pt` may be a single point about which the
    mass moment of inertia is calculated for all regions, or it may be a `dict`
    with one item per region indicating the reference point (value) to use for 
    that region (key). If any region identifier does not appear as a `dict` key,
    an error is raise. Extra keys in the `dict` that do not correspond to any 
    region identifiers are ignored with a warning.

    The mass moment of inertia can be normalized to provide a compactness 
    measure using the formula from Fan, et al. (2015):
    
    .. math::
        C_{NMMI} = \\frac{M A}{2 \\pi I_M}
    
    where :math:`I_M` is the mass moment of inertia of the shape. This 
    represents the ratio of the mass moment of inertia of a circle with the 
    same area and mass as the shape to the mass moment of inertia of the shape.
    In this case, the measure can exceed 1, as will happen when mass is 
    concentrated near the centroid of the shape. This can be interpreted as a 
    more compact distribution than a uniform circle.
    
    References
    ----------
    .. [1] Weaver, James B., and Sidney W. Hess. 1963. "A Procedure for Nonpartisan 
        Districting: Development of Computer Techniques." Yale Law Journal 73 (2): 288–308. 
        https://doi.org/10.2307/794769
    .. [2] Li, Wenwen, Michael F. Goodchild, and Richard Church. 2013. "An Efficient Measure 
        of Compactness for Two-Dimensional Shapes and Its Application in Regionalization 
        Problems." International Journal of Geographical Information Science 27 (6): 1227–50. 
        https://doi.org/10.1080/13658816.2012.752093.
    .. [3] Fan, Chao, Wenwen Li, Levi J. Wolf and Soe W. Myint. 2015 "A Spatiotemporal
        Compactness Pattern Analysis of Congressional Districts to Assess Partisan 
        Gerrymandering: A Case Study with California and North Carolina." Annals of the 
        Association of American Geographers 105 (4): 736-753. 
        https://doi.org/10.1080/00045608.2015.1039109
    """

    ga = _cast(collection)
    ga = orient_polygons(ga) # shapely.orient_polygons(ga)

    # Handle weights (masses).
    if weights is not None:

        # If weights is a string, interpret as column name and extract from GeoDataFrame
        if isinstance(weights, str):

            if not isinstance(collection, geopandas.GeoDataFrame) or weights not in collection.columns:
                msg = "If `weights` is a string, it must be the name of a column in `collection`, which must be a GeoDataFrame."
                raise ValueError(msg)
            weights = numpy.asarray(collection[weights])
        else:
            weights = numpy.asarray(weights)

    # If regions is missing, calculate MOI per geometry. Must be handled differently 
    # depending on whether weights is missing or normalization is requested.
    if regions is None:
        if weights is None or normalize:
            # If weights are missing, this reduces to second moment of area.
            # If weights are present but we are normalizing, this also reduces 
            # to (normalized) second moment of area. Pass to moment_of_inertia
            # and return.
            return moment_of_inertia(collection, normalize=normalize, ref_pt=ref_pt)
        else: # Weights are present but we are not normalizing
            # Adjust second moment of area for mass. I_M = I_A * m / A
            return moment_of_inertia(collection, normalize=False, ref_pt=ref_pt) * weights / numpy.asarray(shapely.area(ga))

    # Handle region IDs.
    else:

        # If regions is a string, interpret as column name and extract from GeoDataFrame
        if isinstance(regions, str):

            if not isinstance(collection, geopandas.GeoDataFrame) or regions not in collection.columns:
                msg = "If `regions` is a string, it must be the name of a column in `collection`, which must be a GeoDataFrame."
                raise ValueError(msg)
            regions = numpy.asarray(collection[regions])
        else:
            regions = numpy.asarray(regions)

        unique_regions = numpy.unique(regions)

    # If we get to this point, regions are present, but weights might be none. Will
    # be handled differently in loop on unique regions.

    # Handle reference point(s), if provided
    if ref_pt is not None:
        if isinstance(ref_pt, dict):
            k = list(ref_pt.keys())
        
            # Make sure we have one entry per region
            if set(unique_regions) <= set(k):
                if set(unique_regions) < set (k):
                    # Extra unused regions in dict. Issue warning and remove them
                    msg = "Keys found in `ref_pt` that are not regions in `regions`. Extra regions will be ignored."
                    warnings.warn(msg, UserWarning, stacklevel=2)
                # Cast all points in dict values
                ref_pt = {key: _cast_pts_as_array(value) for key, value in ref_pt.items()}

            else:
                msg = "Regions found in `regions` that are not in `ref_pt`. If `ref_pt` is a `dict`, every region must have an entry in the `dict`."
                raise ValueError(msg)
        else:
            ref_pt = _cast_pts_as_array(ref_pt)
            # If not passed as a dictionary, this should be single global reference point
            if ref_pt.shape != (2,):
                msg = f"`ref_pt` must be a single point (or coordinate pair) or a dictionary with one point per region."
                raise ValueError(msg)            
            
    Js = []

    for region in unique_regions:
        # Use mask to calculate moment of inertia over each region. When user
        # omits and `regions`, this effectively becomes a simple
        # loop over ga, returning mass moment of inertia per geometry.
        mask = regions == region
        geoms = ga[mask]

        if weights is None:
            # Calcluate area moment of inertia per region
                        # Determine reference point for shifting, or use centroid
            if ref_pt is None:
                # Use centroid
                pt = None
            elif isinstance(ref_pt, dict):
                # Use regional reference point
                pt = ref_pt[region]
            else:
                # Use global reference point
                pt = ref_pt

            Js.append(moment_of_inertia_global(geoms, normalize=normalize, ref_pt=pt))

        else:
            m = weights[mask]

            # A_tot, Cx, Cy, Ixx, Iyy, J = _moments_about_centroid(geoms)
            moments = numpy.asarray([_moments_about_centroid(geom) for geom in geoms])
            a = moments[:,0]
            cx = moments[:,1]
            cy = moments[:,2]
            c = numpy.column_stack((cx, cy))
            Ixx = moments[:,3]
            Iyy = moments[:,4]
            J = moments[:,5]

            # Area and centroid of region
            A = numpy.sum(a)
            C = numpy.sum(m[:, None] * c, axis=0) / m.sum()

            # Determine reference point for shifting, or use centroid
            if ref_pt is None:
                # Use centroid
                pt = C
            elif isinstance(ref_pt, dict):
                # Use regional reference point
                pt = ref_pt[region]
            else:
                # Use global reference point
                pt = ref_pt

            # Distance squared, don't actually need distance, so don't bother taking square root
            d2 = numpy.sum((c - pt)**2, axis=1)

            J = numpy.sum((m / a) * J + m * d2)

            if normalize:
                J = m.sum() * A / (2 * numpy.pi * J)

            Js.append(J)

    return pandas.Series(Js, index=unique_regions)

def moa_ratio(collection):
    """
    Computes the ratio of the second moment of area (like Li et al (2013)) to
    the moment of area of a circle with the same perimeter.
    """
    msg = "`moa_ratio` is deprecated and will be removed in a future version."
    warnings.warn(msg, DeprecationWarning, stacklevel=2)

    ga = _cast(collection)
    r = shapely.length(ga) / (2 * numpy.pi)
    return (numpy.pi * 0.5 * r**4) / second_areal_moment(ga)

def nmi(collection):
    """
    Computes the normalized moment of inertia

    Notes
    -----
    Implemented as `moment_of_inertia(collection, normalize=True, ref_pt=None)`.
    See `moment_of_inertia` for details.
    """
    
    return moment_of_inertia(collection, normalize=True, ref_pt=None)

def moment_of_inertia_global(collection, normalize=False, ref_pt=None):
    """
    Compute moment of inertia (second moment of area) for an entire collection of geometries combined.

    Parameters
    ----------
    collection : GeoSeries, GeoDataFrame, numpy.ndarray, list
        Input collection of polygons or multipolygons.
    normalize : bool, optional
        If True, returns moment normalized by reference cricle of same area. Default is False.
    ref_pt : GeoSeries, Shapely Point, or array-like of shape (2,), optional
        If provided, shifts moment to be with respect to this point.
        The default behavior (default: ``None``) is to calculate the moment 
        about the centroid of the entire `collection`. Point may be passed as
        as array-like of coordinates or a point geometry (which can include a
        GeoSeries of length 1). To return moment about the origin, explicitly set to (0, 0).

    Returns
    -------
    float
        Moment of inertia for the entire collection.

    Notes
    -------
    This is a convenience function to calculate the second moment of area for
    an entire collection, which will usually be faster than running a geospatial
    dissolve on the geometries and then running `moment_of_inertia` on the 
    result.

    This will *not* calculate a mass moment of inertia. To calculate the mass
    moment of inertia for an entire collection, weighted by the masses of the 
    geometries in the collection, assign all geometries to the same region:

    `moment_of_inertia_regions(collection, regions=numpy.repeat(1, len(collection))), weights=<weights vector>`

    The `normalize` and `ref_pt` parameters may be used as usual.
    """
    ga = _cast(collection)
    ga = orient_polygons(ga) # shapely.orient_polygons(ga)

    A, Cx, Cy, Ixx, Iyy, J = _moments_about_centroid(ga)

    if ref_pt is not None:
        dx = Cx - ref_pt[0]
        dy = Cy - ref_pt[1]
        J += A * (dx**2 + dy**2)

    if normalize:
        J = A**2 / (2 * numpy.pi * J)

    return J

# -------------------------
# Helper Functions for Inertial Measures
# -------------------------

def _dump_rings(geoms):
    """
    Yield all exterior and interior rings of a collection of polygons/multipolygons
    as numpy arrays of coordinates.

    Parameters
    ----------
    geoms : sequence of shapely geometries
        Polygons or multipolygons.

    Yields
    ------
    numpy.ndarray
        Array of shape (n, 2) containing coordinates of a ring.
    """
    for poly in shapely.get_parts(geoms):
        yield shapely.get_coordinates(poly.exterior)
        for interior in poly.interiors:
            yield shapely.get_coordinates(interior)

@njit
def _geometric_moments_ring(pts, shift_to_centroid=True):
    """
    Compute area, centroid, and second moments of a single polygon ring.

    Parameters
    ----------
    pts : numpy.ndarray
        Array of coordinates defining the ring, shape (n, 2).
    shift_to_centroid : bool, default True
        If True, apply the parallel axis theorem to shift moments to the ring centroid.

    Returns
    -------
    A : float
        Signed area of the ring.
    cx, cy : float
        Coordinates of the ring centroid.
    Ixx, Iyy : float
        Second moments of area about the centroid (or origin if shift_to_centroid=False).

    Notes
    -----
    - Polar moment of area J can be obtained by summing Ixx + Iyy at the calling level.
    - This function does not compute J itself, leaving derived quantities to higher-level functions.
    """
    x = pts[:, 0]
    y = pts[:, 1]

    cross = x[:-1] * y[1:] - x[1:] * y[:-1]

    A = cross.sum() / 2
    Sx = ((x[:-1] + x[1:]) * cross).sum()
    Sy = ((y[:-1] + y[1:]) * cross).sum()

    Ixx_origin = ((y[:-1]**2 + y[:-1]*y[1:] + y[1:]**2) * cross).sum() / 12
    Iyy_origin = ((x[:-1]**2 + x[:-1]*x[1:] + x[1:]**2) * cross).sum() / 12

    cx = Sx / (6 * A)
    cy = Sy / (6 * A)

    if shift_to_centroid:
        Ixx = Ixx_origin - A * cy**2
        Iyy = Iyy_origin - A * cx**2
    else:
        Ixx, Iyy = Ixx_origin, Iyy_origin

    return A, cx, cy, Ixx, Iyy


def _moments_about_centroid(geoms):
    """
    Compute combined moments of area for a collection of polygons.

    Parameters
    ----------
    geoms : sequence of shapely geometries
        Input polygons or multipolygons.

    Returns
    -------
    A_tot : float
        Total area of the collection.
    Cx, Cy : float
        Centroid coordinates of the combined geometry collection.
    Ixx, Iyy : float
        Second moments of area about the centroid.
    J : float
        Polar moment of area (Ixx + Iyy) about the centroid.
    """
    A_tot = 0.0
    Mx_tot = 0.0
    My_tot = 0.0
    Ixx0_tot = 0.0
    Iyy0_tot = 0.0

    for ring in _dump_rings(geoms):
        A, cx, cy, Ixx_c, Iyy_c = _geometric_moments_ring(ring, shift_to_centroid=False)

        # First moments
        A_tot += A
        Mx_tot += A * cx
        My_tot += A * cy

        # Accumulate inertia about origin
        Ixx0_tot += Ixx_c
        Iyy0_tot += Iyy_c

    # Centroid of entire collection
    Cx = Mx_tot / A_tot
    Cy = My_tot / A_tot

    # Shift to centroid
    Ixx = Ixx0_tot - A_tot * Cy**2
    Iyy = Iyy0_tot - A_tot * Cx**2

    J = Ixx + Iyy

    return A_tot, Cx, Cy, Ixx, Iyy, J

# -------------------- OTHER MEASURES -------------------- #


def reflexive_angle_ratio(collection):
    """
    The Taylor reflexive angle index, measure OS_4 in :cite:`altman1998Districting`

    (N-R)/(N+R), the difference in number between non-reflexive angles and
    reflexive angles in a polygon, divided by the number of angles in the
    polygon in general.
    """
    angles, geom_indices = get_angles(collection, return_indices=True)
    return (
        pandas.DataFrame.from_dict(dict(is_reflex=angles < 0, geom_ix=geom_indices))
        .groupby("geom_ix")
        .is_reflex.mean()
        .values
    )



import contextlib

import numpy
import pandas
from packaging.version import Version
from scipy.special import entr

with contextlib.suppress(ImportError, ModuleNotFoundError):
    # gets handled in the _cast function.
    import shapely


# from nowosad and stepinski
# https://doi.org/10.1080/13658816.2018.1511794


def _overlay(a, b, return_indices=False):
    """
    Compute geometries from overlaying a onto b
    """
    tree = shapely.STRtree(a)
    bix, aix = tree.query(b)
    overlay = shapely.intersection(a[aix], b[bix])
    if return_indices:
        return aix, bix, overlay
    return overlay


def _cast(collection):
    """
    Cast a collection to a shapely geometry array.
    """
    try:
        import geopandas
    except (ImportError, ModuleNotFoundError) as exception:
        raise type(exception)(
            "shapely and geopandas are required for map comparison statistics."
        ) from None

    if Version(shapely.__version__) < Version("2"):
        raise ImportError("Shapely 2.0 or newer is required.")

    if isinstance(collection, geopandas.GeoSeries | geopandas.GeoDataFrame):
        return numpy.asarray(collection.geometry.array)
    else:
        if isinstance(collection, numpy.ndarray | list):
            return numpy.asarray(collection)
        else:
            return numpy.array([collection])


def external_entropy(a, b, balance=0, base=numpy.e):
    """
    The harmonic mean summarizing the overlay entropy of two
    sets of polygons: a onto b and b onto a.

    Called the v-measure in :cite:`nowosad2018`

    Parameters
    ----------
    a : geometry array of polygons
        array of polygons
    b : geometry array of polygons
        array of polygons
    balance:  float
        weight that describing the relative importance of pattern a or pattern b.
        When large and positive, we weight the measure more to ensure polygons in b
        are fully contained by polygons in a. When large and negative,
        we weight the pattern to ensure polygons in A are fully contained
        by polygons in b. Corresponds to the log of beta in {cite}`Nowosad2018`.
    base: float
        base of logarithm to use throughout computation
    Returns
    --------
    (n,) array expressing the entropy of the areal distributions
    of a's splits by partition b.

    Example
    -------

    >>> r1 = geopandas.read_file('tests/regions.zip', layer='regions1')
    >>> r2 = geopandas.read_file('tests/regions.zip', layer='regions2')
    >>> external_entropy(r1, r2)
    """
    a = _cast(a)
    b = _cast(b)
    beta = numpy.exp(balance)
    aix, bix, ab = _overlay(a, b, return_indices=True)
    a_areas = shapely.area(a)
    b_areas = shapely.area(b)
    ab_areas = shapely.area(ab)
    b_onto_a = _overlay_entropy(aix, a_areas, ab_areas, base=base)  # SjZ
    # SZ, as sabre has entropy.empirical(rowSums(xtab), unit='log2')
    b_onto_a /= areal_entropy(areas=b_areas, local=False, base=base)
    a_onto_b = _overlay_entropy(bix, b_areas, ab_areas, base=base)  # SjR
    # SR, as sabre has entropy.empirical(colSums(xtab), unit='log2')
    a_onto_b /= areal_entropy(areas=a_areas, local=False, base=base)

    c = 1 - numpy.average(b_onto_a, weights=a_areas)
    h = 1 - numpy.average(a_onto_b, weights=b_areas)

    return (1 + beta) * h * c / ((beta * h) + c)


def completeness(a, b, local=False, base=numpy.e):
    """
    The completeness of the partitions of polygons in a to those in a.
    Closer to 1 when all polygons in a are fully contained within polygons in b.
    From :cite:`nowosad2018`

    Parameters
    ----------
    a : geometry array of polygons
        array of polygons
    b : geometry array of polygons
        array of polygons
    local: bool (default: False)
        whether or not to provide local scores for each polygon. If True, the
        completeness for polygons in a are returned.
    scale: bool (default: None)
        whether to scale the completeness score(s). By default, completeness is
        is scaled for local scores so that the average of the local scores is
        the overall map completeness. If not local, then completeness is returned
        unscaled. You can also set local=True and scale=False to get raw components
        of the completeness, whose sum is the completeness for the entire map.
        Global re-scaled scores (local=False & scale=True) are not supported.
    base: bool (default=None)
        what base to use for the entropy calculations. The default is base e,
        which means entropy is measured in "nats."

    Example
    -------

    >>> r1 = geopandas.read_file('tests/regions.zip', layer='regions1')
    >>> r2 = geopandas.read_file('tests/regions.zip', layer='regions2')
    >>> completeness(r1, r2)
    """
    a = _cast(a)
    b = _cast(b)
    ohi = overlay_entropy(a, b, standardize=True, local=True, base=base)
    a_areas = shapely.area(a)
    w = a_areas / a_areas.sum()
    ci = (w * (1 - ohi)) / w.sum()
    if local:
        return ci
    return ci.sum()


def homogeneity(a, b, local=False, base=numpy.e):
    """
    The homogeneity of polygons from a partitioned by b.

    From :cite:`nowosad2018`

    This is equal to completeness(b,a).

    It is closer to 1 when all polygons in b correspond well to polygons in a.

    Parameters
    ----------
    a : geometry array of polygons
        array of polygons
    b : geometry array of polygons
        array of polygons
    local: bool (default: False)
        whether or not to provide local scores for each polygon. If True, the
        homogeneity for polygons in b are returned.
    scale: bool (default: None)
        whether to scale the completeness score(s). By default, completeness is
        is scaled for local scores so that the average of the local scores is
        the overall map completeness. If not local, then completeness is returned
        unscaled. You can also set local=True and scale=False to get raw components
        of the completeness, whose sum is the completeness for the entire map.
        Global re-scaled scores (local=False & scale=True) are not supported.
    base: bool (default=None)
        what base to use for the entropy calculations. The default is base e,
        which means entropy is measured in "nats."

    Example
    -------

    >>> r1 = geopandas.read_file('tests/regions.zip', layer='regions1')
    >>> r2 = geopandas.read_file('tests/regions.zip', layer='regions2')
    >>> homogeneity(r1, r2)
    """
    return completeness(b, a, local=local, base=base)


def overlay_entropy(a, b, standardize=True, local=False, base=numpy.e):
    """
    The entropy of how n zones in a are split by m partitions in b,
    where n is the number of polygons in a and m is the number
    of partitions in b. This is the "overlay entropy", since
    the set of polygons constructed from intersection(a,b) is often
    called the "overlay" of A onto B.


    Larger when zones in a are uniformly split into many even pieces
    by partitions in b, and small when zones in A correspond well to
    zones in B.

    Parameters
    -----------
    a : geometry array of polygons
        a set of polygons (the "target") for whom the areal entropy is calculated
    b : geometry array of polygons
        a set of polygons (the "frame") that splits a

    Returns
    --------
    (n,) array expressing the entropy of the areal distributions
    of a's splits by partition b.

    Example
    -------

    >>> r1 = geopandas.read_file('tests/regions.zip', layer='regions1')
    >>> r2 = geopandas.read_file('tests/regions.zip', layer='regions2')
    >>> overlay_entropy(r1, r2)
    """
    a = _cast(a)
    b = _cast(b)
    aix, bix, ab = _overlay(a, b, return_indices=True)
    a_areas = shapely.area(a)
    b_areas = shapely.area(b)
    h = _overlay_entropy(aix, a_areas, shapely.area(ab), base=base)
    if standardize:
        h /= areal_entropy(None, areas=b_areas, local=False, base=base)
    if local:
        return h
    return h.sum()


def _overlay_entropy(aix, a_areas, ab_areas, base):
    """
    direct function to compute overlay entropies
    """
    mapping = pandas.DataFrame.from_dict(
        dict(
            a=aix,
            area=ab_areas,
            a_area=a_areas[aix],
        )
    )
    mapping["frac"] = mapping.area / mapping.a_area
    mapping["entropy"] = entr(mapping.frac.values) / numpy.log(base)
    result = mapping.groupby("a").entropy.sum().copy()
    result[result < 0] = 0
    return result


def areal_entropy(polygons=None, areas=None, local=False, base=numpy.e):
    """
    Compute the entropy of the distribution of polygon areas.

    Parameters
    ----------
    polygons: numpy array of geometries
        polygons whose distribution of entropies needs to be computed.
        Should not be provided if areas is provided.
    areas: numpy array
        areas to use to compute entropy. SHould not be provided if
        polygons are provided.
    local: bool (default: False)
        whether to return the total entropy of the areal distribution
        (False), or to return the contribution to entropy made
        by each of area (True).

    Returns
    -------
    Total map entropy or (n,) vector of local entropies.

    Example
    -------

    >>> r1 = geopandas.read_file('tests/regions.zip', layer='regions1')
    >>> r2 = geopandas.read_file('tests/regions.zip', layer='regions2')
    >>> areal_entropy(polygons=r1)
    """
    assert not ((polygons is None) & (areas is None)), (
        "Either polygons or precomputed areas must be provided."
    )
    assert not ((polygons is not None) & (areas is not None)), (
        "Only one of polygons or areas should be provided."
    )
    if polygons is None:
        assert areas is not None, "If polygons are not provided, areas should be."
    if areas is None:
        assert polygons is not None, "If areas are not provided, polygons should be."
        polygons = _cast(polygons)
        areas = shapely.area(polygons)
    result = entr(areas / areas.sum()) / numpy.log(base)
    result[result < 0] = 0
    if local:
        return result
    return result.sum()



from libpysal import examples
from libpysal.weights.util import get_points_array
from esda import correlogram
import numpy as np
from numpy.testing import assert_array_almost_equal
from scipy import spatial
import geopandas as gpd
import pytest


sac = gpd.read_file(examples.load_example("Sacramento1").get_path("sacramentot2.shp"))
sac = sac.to_crs(sac.estimate_utm_crs())  # now in meters)

dsupport = [i + 500 for i in range(0, 2000, 500)]
ksupport = list(range(1, 6))

try:
    import statsmodels  # noqa F401
except ImportError:
    statsmodels = None

def test_distance_correlogram():
    corr = correlogram(sac.geometry.centroid, sac.HH_INC, dsupport)

    test_data = np.array(
        [
            0.05822723177762817,
            0.49206877942505206,
            0.45494217612839183,
            0.5625914469490942,
        ]
    )

    assert_array_almost_equal(corr.I, test_data)


def test_k_distance_correlogram():
    corr = correlogram(sac.geometry.centroid, sac.HH_INC, ksupport, distance_type="knn")

    test_data = np.array([0.62411734, 0.59734846, 0.56958116, 0.54252517, 0.54093269])
    assert_array_almost_equal(corr.I, test_data)


def test_unspecified_distances():
    corr = correlogram(sac.geometry.centroid, sac.HH_INC, distance_type="knn")

    assert len(corr) == 50
    known = [
        0.62411734,
        0.48356077,
        0.40204787,
        0.36573375,
        0.32139756,
        -0.00165216,
        -0.00217858,
        -0.00237038,
        -0.00260488,
        -0.00248756,
    ]
    assert_array_almost_equal(  # check the first five and last five
        [*corr.I[:5], *corr.I[-5:]],
        known,
    )

@pytest.mark.skipif(statsmodels is None, reason="lowess requires statsmodels")
def test_lowess_correlogram():
    corr = correlogram(
        sac.geometry.centroid, sac.HH_INC, support=dsupport, statistic="lowess"
    )

    test_data = np.array([0.586032, 0.377228, 0.268441, 0.336877])

    assert_array_almost_equal(corr.lowess, test_data)

@pytest.mark.skipif(statsmodels is None, reason="lowess requires statsmodels")
def test_lowess_precomputed():
    coords = get_points_array(sac.geometry.centroid)
    n_samples = len(coords)
    D = np.zeros((n_samples, n_samples))
    upper_ix = np.triu_indices_from(D)
    lower_ix = np.tril_indices_from(D)

    # diff in x on top, diff in y on bottom
    D[upper_ix] = spatial.distance_matrix(coords[:, 0, None], coords[:, 0, None])[
        upper_ix
    ]
    D[lower_ix] = spatial.distance_matrix(coords[:, 1, None], coords[:, 1, None])[
        lower_ix
    ]

    corr = correlogram(
        sac.geometry.centroid,
        sac.HH_INC,
        statistic="lowess",
        stat_kwargs=dict(metric="precomputed", coordinates=D),
    )

    assert_array_almost_equal(
        corr.lowess.iloc[:5],
        [0.07460634, 0.05469855, 0.02606339, 0.02323869, 0.01068146],
    )



import numpy as np
import pytest
from libpysal import graph
from libpysal.weights.util import lat2W

from ..join_counts import Join_Counts

parametrize_w = pytest.mark.parametrize(
    "w",
    [
        lat2W(4, 4),
        graph.Graph.from_W(lat2W(4, 4)),
    ],
    ids=["W", "Graph"],
)


class TestJoinCounts:
    """Unit test for Join Counts"""

    def setup_method(self):
        self.y = np.ones(16)
        self.y[0:8] = 0

    @parametrize_w
    def test_default(self, w):
        """Test method"""
        np.random.seed(12345)
        jc = Join_Counts(self.y, w)
        np.testing.assert_allclose(jc.bb, 10.0)
        np.testing.assert_allclose(jc.bw, 4.0)
        np.testing.assert_allclose(jc.ww, 10.0)
        np.testing.assert_allclose(jc.autocorr_neg, 4.0)  # jc.bw
        np.testing.assert_allclose(jc.autocorr_pos, 20.0)
        np.testing.assert_allclose(jc.J, 24.0)
        np.testing.assert_allclose(len(jc.sim_bb), 999)
        np.testing.assert_allclose(jc.p_sim_bb, 0.0030000000000000001)
        np.testing.assert_allclose(np.mean(jc.sim_bb), 5.5465465465465469)
        np.testing.assert_allclose(np.max(jc.sim_bb), 10.0)
        np.testing.assert_allclose(np.min(jc.sim_bb), 0.0)
        np.testing.assert_allclose(len(jc.sim_bw), 999)
        np.testing.assert_allclose(jc.p_sim_bw, 1.0)
        np.testing.assert_allclose(np.mean(jc.sim_bw), 12.811811811811811)
        np.testing.assert_allclose(np.max(jc.sim_bw), 24.0)
        np.testing.assert_allclose(np.min(jc.sim_bw), 7.0)
        np.testing.assert_allclose(8.166666666666666, jc.chi2)
        np.testing.assert_allclose(0.004266724822176128, jc.chi2_p)
        np.testing.assert_allclose(0.008, jc.p_sim_chi2)
        np.testing.assert_allclose(1.0, jc.p_sim_autocorr_neg)
        np.testing.assert_allclose(0.001, jc.p_sim_autocorr_pos)
        np.testing.assert_allclose(0.2653504320039377, jc.sim_autocorr_chi2)

    @parametrize_w
    def test_by_col(self, w):
        import pandas as pd

        df = pd.DataFrame(self.y, columns=["y"])
        np.random.seed(12345)
        r1 = Join_Counts.by_col(
            df, ["y"], w=w, permutations=999
        )  # outvals = ['bb', 'bw', 'ww', 'p_sim_bw', 'p_sim_bb']

        bb = np.unique(r1.y_bb.values)
        bw = np.unique(r1.y_bw.values)
        bb_p = np.unique(r1.y_p_sim_bb.values)
        bw_p = np.unique(r1.y_p_sim_bw.values)
        np.random.seed(12345)
        c = Join_Counts(self.y, w, permutations=999)
        np.testing.assert_allclose(bb, c.bb)
        np.testing.assert_allclose(bw, c.bw)
        np.testing.assert_allclose(bb_p, c.p_sim_bb)
        np.testing.assert_allclose(bw_p, c.p_sim_bw)



import numpy as np


def fdr(pvalues, alpha=0.05):
    """
    Calculate the p-value cut-off to control for
    the false discovery rate (FDR) for multiple testing.

    If by controlling for FDR, all of n null hypotheses
    are rejected, the conservative Bonferroni bound (alpha/n)
    is returned instead.

    Parameters
    ----------
    pvalues     : array
                  (n, ), p values for n multiple tests.
    alpha       : float, optional
                  Significance level. Default is 0.05.

    Returns
    -------
                : float
                  Adjusted criterion for rejecting the null hypothesis.
                  If by controlling for FDR, all of n null hypotheses
                  are rejected, the conservative Bonferroni bound (alpha/n)
                  is returned.

    Notes
    -----

    For technical details see :cite:`Benjamini:2001` and
    :cite:`Castro:2006tz`.


    Examples
    --------
    >>> import libpysal
    >>> import numpy as np
    >>> np.random.seed(10)
    >>> w = libpysal.io.open(libpysal.examples.get_path("stl.gal")).read()
    >>> f = libpysal.io.open(libpysal.examples.get_path("stl_hom.txt"))
    >>> y = np.array(f.by_col['HR8893'])
    >>> from esda.moran import Moran_Local
    >>> from esda import fdr
    >>> lm = Moran_Local(y, w, transformation = "r", permutations = 999)
    >>> fdr(lm.p_sim, 0.1)
    0.002564102564102564
    >>> fdr(lm.p_sim, 0.05) #return the conservative Bonferroni bound
    0.000641025641025641

    """

    n = len(pvalues)
    p_sort = np.sort(pvalues)[::-1]
    index = np.arange(n, 0, -1)
    p_fdr = index * alpha / n
    search = p_sort < p_fdr
    sig_all = np.where(search)[0]
    if len(sig_all) == 0:
        return alpha / n
    else:
        return p_fdr[sig_all[0]]



import numpy
import geodatasets
import geopandas
import pytest
from libpysal.weights import Rook
from libpysal.graph import Graph
from sklearn.linear_model import TheilSenRegressor
from esda.moran_local_mv import MoranLocalPartial, MoranLocalConditional
from esda.moran import Moran_Local_BV

def rsrook(df):
    w_classic = Rook.from_dataframe(df) 
    w_classic.transform = 'r'
    return w_classic

@pytest.fixture(scope='module')
def data():
    df = geopandas.read_file(geodatasets.get_path("geoda.lansing1"))
    df = df[df.FIPS.str.match("2606500[01234]...") | (df.FIPS == "26065006500")]
    df=df.reset_index(drop=True)
    y = df.HH_INC.values.reshape(-1,1)
    X = df.HSG_VAL.values.reshape(-1,1)
    yield y,X,df

@pytest.fixture(scope='module', 
                params = [
                        rsrook,
                        lambda df: Graph.build_contiguity(df.reset_index(drop=True)).transform('r')
                        ],
                ids=['W', 'Graph']
                )
def graph(data, request):
    _,_,df = data
    return request.param(df)

def test_partial_runs(data, graph):
    """Check if the class computes successfully in a default configuration"""
    y,X,df = data
    m = MoranLocalPartial(permutations=1).fit(X,y,graph)
    # done, just check if it runs

def test_partial_accuracy(data, graph):
    """Check if the class outputs expected results at a given seed"""
    y,X,df = data
    numpy.random.seed(111221)
    m = MoranLocalPartial(permutations=10).fit(X,y,graph)
    # compute result by hand
    zy = (y - y.mean())/y.std()
    wz = (graph.sparse @ zy)
    zx = (X - X.mean(axis=0))/X.std(axis=0)
    rho = numpy.corrcoef(zy.squeeze(), zx.squeeze())[0,1]
    left = zy - rho * zx
    scale = (graph.n-1) / (graph.n * (1 - rho**2))
    # (y - rho x)*wy
    manual = (left*wz).squeeze() * scale
    # check values
    numpy.testing.assert_allclose(manual, m.association_)

    # check significances are about 18
    print(m.significance_)
    numpy.testing.assert_allclose((m.significance_ < .4).sum(), 26, atol=1)
    numpy.testing.assert_equal((m.significance_[:5] < .4), [True, True, True, False, False])

    # check quad
    is_cluster = numpy.prod(m.partials_, axis=1) >= 0
    is_odd_label = m.labels_ % 2
    numpy.testing.assert_equal(is_cluster, is_odd_label)

def test_partial_unscaled(data, graph):
    """Check if the variance scaling behaves as expected"""
    y,X,df = data
    m = MoranLocalPartial(permutations=0, unit_scale=True).fit(X,y,graph)
    m2 = MoranLocalPartial(permutations=0, unit_scale=False).fit(X,y,graph)
    # variance in the partials_ should be different
    s1y,s1x = m.partials_.std(axis=0)
    s2y,s2x = m2.partials_.std(axis=0)
    assert s1y > s2y, "variance is incorrectly scaled for y"
    assert s1x < s2x, "variance is incorrectly scaled for x"

def test_partial_uvquads(data, graph):
    """Check that the quadrant decisions vary correctly with the inputs"""
    y,X,df = data
    m = MoranLocalPartial(permutations=0, partial_labels=False).fit(X,y,graph)
    bvx = Moran_Local_BV(X,y,graph,permutations=0)
    numpy.testing.assert_array_equal(m.labels_, bvx.q)

def test_conditional_runs(data, graph):
    """Check that the class completes successfully in a default configuration"""
    y,X,df = data
    a = MoranLocalConditional(permutations=1).fit(X,y,graph)
    #done, just check if it runs

def test_conditional_accuracy(data, graph):
    """Check that the class outputs expected values for a given seed"""
    y,X,df = data
    numpy.random.seed(111221)
    a = MoranLocalConditional(permutations=10).fit(X,y,graph)

    # compute result by hand
    zy = (y - y.mean())/y.std()
    wz = (graph.sparse @ zy)
    zx = (X - X.mean(axis=0))/X.std(axis=0)
    wzx = graph.sparse @ zx
    rho = numpy.corrcoef(zy.squeeze(), zx.squeeze())[0,1]
    mean = zy * wz - rho * zx * wz - rho * zy * wzx + rho**2 * zx * wzx
    scale = (graph.n-1) / (graph.n * (1 - rho**2))
    
    manual = numpy.asarray(mean * scale).squeeze()
    # check values, may not be identical because of the 
    # matrix inversion least squares estimator used in scikit
    numpy.testing.assert_allclose(manual, a.association_)

    # check significances
    numpy.testing.assert_equal((a.significance_ < .4).sum(), 31)
    numpy.testing.assert_equal((a.significance_[:5] < .4), [False, False, True, True, False])

    is_cluster = numpy.prod(a.partials_, axis=1) >= 0
    is_odd_label = (a.labels_ % 2).astype(bool)
    numpy.testing.assert_equal(is_cluster, is_odd_label)

def test_conditional_unscaled(data, graph):
    """Check that the variance scaling behaves as expected"""
    y,X,df = data
    a = MoranLocalConditional(permutations=0, unit_scale=True).fit(X,y,graph)
    a2 = MoranLocalConditional(permutations=0, unit_scale=False).fit(X,y,graph)
    assert (a.partials_.std(axis=0) < a2.partials_.std(axis=0)).all(), (
        "variance is not scaled correctly in partial regression."
        )

def test_conditional_transformer(data, graph):
    """Check that an alternative regressor can be used to calculate y|X"""
    y,X, df = data
    a = MoranLocalConditional(permutations=0, transformer=TheilSenRegressor).fit(X,y,graph)
    # done, should just complete


import numpy as np
import pytest

from esda.crand import vec_permutations


def test_vec_permutations_basic():
    """Test vec_permutations with basic input."""
    result = vec_permutations(5, 24, 10, seed=12345)
    assert len(result) == 10, "Should create 10 permutations"
    assert all(isinstance(item, np.ndarray) for item in result), (
        "Each permutation should be an array"
    )
    expected = np.array(
        [
            [10, 19, 8, 17, 3],
            [21, 16, 8, 15, 10],
            [18, 4, 20, 3, 13],
            [4, 9, 20, 7, 15],
            [16, 15, 4, 21, 0],
            [9, 10, 8, 1, 11],
            [20, 11, 6, 15, 0],
            [10, 3, 11, 14, 16],
            [12, 6, 8, 7, 19],
            [18, 19, 4, 13, 15],
        ]
    )
    np.testing.assert_array_equal(result, expected)



# based off: https://github.com/pysal/esda/blob/master/esda/tests/test_join_counts.py
import numpy as np
import pytest
from libpysal import graph
from libpysal.weights.util import lat2W

from esda.join_counts_local import Join_Counts_Local

parametrize_w = pytest.mark.parametrize(
    "w",
    [
        lat2W(4, 4),
        graph.Graph.from_W(lat2W(4, 4)),
    ],
    ids=["W", "Graph"],
)


class TestJoinCountsLocal:
    """Unit test for Local Join Counts (univariate)"""

    def setup_method(self):
        self.y = np.ones(16)
        self.y[0:8] = 0

    @parametrize_w
    def test_defaults(self, w):
        np.random.seed(12345)
        ljc = Join_Counts_Local(connectivity=w).fit(self.y)
        assert np.array_equal(ljc.LJC, [0, 0, 0, 0, 0, 0, 0, 0, 2, 3, 3, 2, 2, 3, 3, 2])



import libpysal
import numpy as np
import pytest

from .. import moran, util


class TestFdr:
    def setup_method(self):
        self.w = libpysal.io.open(libpysal.examples.get_path("stl.gal")).read()
        f = libpysal.io.open(libpysal.examples.get_path("stl_hom.txt"))
        self.y = np.array(f.by_col["HR8893"])

    def test_fdr(self):
        lm = moran.Moran_Local(
            self.y, self.w, transformation="r", permutations=999, seed=10
        )
        assert pytest.approx(util.fdr(lm.p_sim, 0.1)) == 0.002564102564102564
        assert pytest.approx(util.fdr(lm.p_sim, 0.05)) == 0.001282051282051282



import libpysal
import numpy as np
import pytest

from esda.geary_local import Geary_Local

parametrize_w = pytest.mark.parametrize(
    "w",
    [
        libpysal.io.open(libpysal.examples.get_path("stl.gal")).read(),
        libpysal.graph.Graph.from_W(
            libpysal.io.open(libpysal.examples.get_path("stl.gal")).read()
        ),
    ],
    ids=["W", "Graph"],
)


class TestGearyLocal:
    def setup_method(self):
        np.random.seed(10)
        f = libpysal.io.open(libpysal.examples.get_path("stl_hom.txt"))
        self.y = np.array(f.by_col["HR8893"])

    @parametrize_w
    def test_defaults(self, w):
        lG = Geary_Local(connectivity=w).fit(self.y)
        np.testing.assert_allclose(lG.localG[0], 0.696703432)
        np.testing.assert_allclose(lG.p_sim[0], 0.19)



import libpysal
import numpy as np
import pytest
from libpysal.common import ATOL, RTOL
from libpysal.weights.distance import DistanceBand

from .. import getisord

POINTS = np.array([(10, 10), (20, 10), (40, 10), (15, 20), (30, 20), (30, 30)])

parametrize_w = pytest.mark.parametrize(
    "w",
    [
        DistanceBand(POINTS, threshold=15),
        libpysal.graph.Graph.from_W(DistanceBand(POINTS, threshold=15)),
    ],
    ids=["W", "Graph"],
)


class TestGetisG:
    def setup_method(self):
        self.y = np.array([2, 3, 3.2, 5, 8, 7])
        np.random.seed(10)

    @parametrize_w
    def test_default(self, w):
        g = getisord.G(self.y, w)
        np.testing.assert_allclose(g.G, 0.55709779, rtol=RTOL, atol=ATOL)
        np.testing.assert_allclose(g.p_norm, 0.172936, rtol=RTOL, atol=ATOL)

    @parametrize_w
    def test_by_col(self, w):
        import pandas as pd

        df = pd.DataFrame(self.y, columns=["y"])
        np.random.seed(12345)
        r1 = getisord.G.by_col(df, ["y"], w=w)
        this_getisord = np.unique(r1.y_g.values)
        this_pval = np.unique(r1.y_p_sim.values)
        np.random.seed(12345)
        stat = getisord.G(self.y, w)
        np.testing.assert_allclose(this_getisord, stat._statistic)
        np.testing.assert_allclose(this_pval, stat.p_sim)


class TestGLocal:
    def setup_method(self):
        self.y = np.array([2, 3, 3.2, 5, 8, 7])
        np.random.seed(10)

    @parametrize_w
    def test_binary(self, w):
        lg = getisord.G_Local(self.y, w, transform="B", seed=10)
        np.testing.assert_allclose(lg.Zs[0], -1.0136729, rtol=RTOL, atol=ATOL)
        np.testing.assert_allclose(lg.p_sim[0], 0.102, rtol=RTOL, atol=ATOL)
        np.testing.assert_allclose(lg.p_z_sim[0], 0.153923, rtol=RTOL, atol=ATOL)

    @parametrize_w
    def test_row_standardized(self, w):
        lg = getisord.G_Local(self.y, w, transform="R", seed=10)
        np.testing.assert_allclose(lg.Zs[0], -0.62074534, rtol=RTOL, atol=ATOL)
        np.testing.assert_allclose(lg.p_sim[0], 0.102, rtol=RTOL, atol=ATOL)

    @parametrize_w
    def test_star_binary(self, w):
        lg = getisord.G_Local(self.y, w, transform="B", star=True, seed=10)
        np.testing.assert_allclose(lg.Zs[0], -1.39727626, rtol=RTOL, atol=ATOL)
        np.testing.assert_allclose(lg.p_sim[0], 0.102, rtol=RTOL, atol=ATOL)

    @parametrize_w
    @pytest.mark.xfail(
        reason="Intermittently does not warn for W param; reason unknown; see gh#331"
    )
    def test_star_row_standardized_warning(self, w):
        with pytest.warns(UserWarning, match="Gi\\* requested, but"):
            getisord.G_Local(self.y, w, transform="R", star=True, seed=10)

    @parametrize_w
    def test_star_row_standardized_values(self, w):
        lg = getisord.G_Local(self.y, w, transform="R", star=True, seed=10)
        np.testing.assert_allclose(lg.Zs[0], -0.62488094, rtol=RTOL, atol=ATOL)
        np.testing.assert_allclose(lg.p_sim[0], 0.102, rtol=RTOL, atol=ATOL)

    @parametrize_w
    def test_by_col(self, w):
        import pandas as pd

        df = pd.DataFrame(self.y, columns=["y"])
        r1 = getisord.G_Local.by_col(df, ["y"], w=w, seed=12345)
        stat = getisord.G_Local(self.y, w, seed=12345)
        np.testing.assert_allclose(r1.y_g_local.values, stat.Gs)
        np.testing.assert_allclose(r1.y_p_sim, stat.p_sim)



import numpy as np
import pandas as pd
from libpysal import weights
from sklearn.base import BaseEstimator

from esda.crand import _prepare_univariate
from esda.crand import crand as _crand_plus
from esda.crand import njit as _njit


class Geary_Local(BaseEstimator):  # noqa: N801
    """Local Geary - Univariate"""

    def __init__(
        self,
        connectivity=None,
        labels=False,
        sig=0.05,
        permutations=999,
        n_jobs=1,
        keep_simulations=True,
        seed=None,
        island_weight=0,
        drop_islands=True,
    ):
        """
        Initialize a Local_Geary estimator

        Parameters
        ----------
        connectivity     : W | Graph
                           spatial weights instance as W or Graph aligned with y
        labels           : boolean
                           (default=False)
                           If True use, label if an observation
                           belongs to an outlier, cluster, other,
                           or non-significant group. 1 = outlier,
                           2 = cluster, 3 = other, 4 = non-significant.
                           Note that this is not the exact same as the
                           cluster map produced by GeoDa.
        sig              : float
                           (default=0.05)
                           Default significance threshold used for
                           creation of labels groups.
        permutations     : int
                           (default=999)
                           number of random permutations for calculation
                           of pseudo p_values
        n_jobs           : int
                           (default=1)
                           Number of cores to be used in the conditional
                           randomisation. If -1, all available cores are used.
        keep_simulations : Boolean
                           (default=True)
                           If True, the entire matrix of replications under
                           the null is stored in memory and accessible;
                           otherwise, replications are not saved
        seed             : None/int
                           Seed to ensure reproducibility of conditional
                           randomizations. Must be set here, and not outside
                           of the function, since numba does not correctly
                           interpret external seeds nor
                           numpy.random.RandomState instances.
        island_weight :
            value to use as a weight for the "fake" neighbor for every island.
            If numpy.nan, will propagate to the final local statistic depending
            on the `stat_func`. If 0, then the lag is always zero for islands.
        drop_islands : bool (default True)
            Whether or not to preserve islands as entries in the adjacency
            list. By default, observations with no neighbors do not appear
            in the adjacency list. If islands are kept, they are coded as
            self-neighbors with zero weight. See ``libpysal.weights.to_adjlist()``.

        Attributes
        ----------
        localG          : numpy array
                          array containing the observed univariate
                          Local Geary values.
        p_sim           : numpy array
                          array containing the simulated
                          p-values for each unit.
        labs            : numpy array
                          array containing the labels for if each observation.
        """

        self.connectivity = connectivity
        self.labels = labels
        self.sig = sig
        self.permutations = permutations
        self.n_jobs = n_jobs
        self.keep_simulations = keep_simulations
        self.seed = seed
        self.island_weight = island_weight
        self.drop_islands = drop_islands

    def fit(self, x):
        """
        Parameters
        ----------
        x                : numpy.ndarray
                           array containing continuous data

        Returns
        -------
        the fitted estimator.

        Notes
        -----
        Technical details and derivations can be found in :cite:`Anselin1995`.

        Examples
        --------
        Guerry data replication GeoDa tutorial
        >>> import libpysal as lp
        >>> import geopandas as gpd
        >>> guerry = lp.examples.load_example('Guerry')
        >>> guerry_ds = gpd.read_file(guerry.get_path('Guerry.shp'))
        >>> w = libpysal.weights.Queen.from_dataframe(guerry_ds)
        >>> y = guerry_ds['Donatns']
        >>> lG = Local_Geary(connectivity=w).fit(y)
        >>> lG.localG[0:5]
        >>> lG.p_sim[0:5]
        """
        x = np.asarray(x).flatten()

        w = self.connectivity
        if isinstance(w, weights.W):
            w.transform = "r"
        else:
            w = w.transform("r")

        permutations = self.permutations
        sig = self.sig
        keep_simulations = self.keep_simulations
        n_jobs = self.n_jobs

        self.localG = self._statistic(x, w, self.drop_islands)

        if permutations:
            self.p_sim, self.rlocalG = _crand_plus(
                z=(x - np.mean(x)) / np.std(x),
                w=w,
                observed=self.localG,
                permutations=permutations,
                keep=keep_simulations,
                n_jobs=n_jobs,
                stat_func=_local_geary,
                island_weight=self.island_weight,
            )

        if self.labels:
            Eij_mean = np.mean(self.localG)
            x_mean = np.mean(x)
            # Create empty vector to fill
            self.labs = np.empty(len(x)) * np.nan
            # Outliers
            locg_lt_eij = self.localG < Eij_mean
            p_leq_sig = self.p_sim <= sig
            self.labs[locg_lt_eij & (x > x_mean) & p_leq_sig] = 1
            # Clusters
            self.labs[locg_lt_eij & (x < x_mean) & p_leq_sig] = 2
            # Other
            self.labs[(self.localG > Eij_mean) & p_leq_sig] = 3
            # Non-significant
            self.labs[self.p_sim > sig] = 4

        return self

    @staticmethod
    def _statistic(x, w, drop_islands):
        # Caclulate z-scores for x
        zscore_x = (x - np.mean(x)) / np.std(x)
        # Create focal (xi) and neighbor (zi) values
        if isinstance(w, weights.W):
            adj_list = w.to_adjlist(remove_symmetric=False, drop_islands=drop_islands)
            zseries = pd.Series(zscore_x, index=w.id_order)
        else:
            adj_list = w.adjacency
            if drop_islands:
                adj_list = adj_list.drop(w.isolates)
            adj_list = adj_list.reset_index()
            zseries = pd.Series(zscore_x, index=w.unique_ids)
        zi = zseries.loc[adj_list.focal].values
        zj = zseries.loc[adj_list.neighbor].values
        # Carry out local Geary calculation
        gs = adj_list.weight.values * (zi - zj) ** 2
        # Reorganize data
        adj_list_gs = pd.DataFrame(adj_list.focal.values, gs).reset_index()
        adj_list_gs.columns = ["gs", "ID"]
        adj_list_gs = adj_list_gs.groupby(by="ID").sum()
        if isinstance(w, weights.W):
            # Rearrange data based on w id order
            adj_list_gs["w_order"] = w.id_order
            adj_list_gs.sort_values(by="w_order", inplace=True)

        localG = adj_list_gs.gs.values

        return localG


# --------------------------------------------------------------
# Conditional Randomization Function Implementations
# --------------------------------------------------------------

# Note: does not using the scaling parameter


@_njit(fastmath=True)
def _local_geary(i, z, permuted_ids, weights_i, scaling):  # noqa: ARG001
    other_weights = weights_i[1:]
    zi, zrand = _prepare_univariate(i, z, permuted_ids, other_weights)
    return (zi - zrand) ** 2 @ other_weights



import libpysal
import numpy as np
import pytest
from libpysal.common import ATOL, RTOL, pandas
from libpysal.weights.distance import KNN, Kernel

from .. import smoothing as sm

PANDAS_EXTINCT = pandas is None


class TestFlatten:
    def setup_method(self):
        self.input = [[1, 2], [3, 3, 4], [5, 6]]

    def test_flatten(self):
        out1 = sm.flatten(self.input)
        out2 = sm.flatten(self.input, unique=False)
        assert out1 == [1, 2, 3, 4, 5, 6]
        assert out2 == [1, 2, 3, 3, 4, 5, 6]


class TestWMean:
    def setup_method(self):
        self.d = np.array([5, 4, 3, 1, 2])
        self.w1 = np.array([10, 22, 9, 2, 5])
        self.w2 = np.array([10, 14, 17, 2, 5])

    def test_weighted_median(self):
        out1 = sm.weighted_median(self.d, self.w1)
        out2 = sm.weighted_median(self.d, self.w2)
        assert out1 == 4
        assert out2 == 3.5


class TestAgeStd:
    def setup_method(self):
        self.e = np.array([30, 25, 25, 15, 33, 21, 30, 20])
        self.b = np.array([1000, 1000, 1100, 900, 1000, 900, 1100, 900])
        self.s_e = np.array([100, 45, 120, 100, 50, 30, 200, 80])
        self.s_b = np.array([1000, 900, 1000, 900, 1000, 900, 1000, 900])
        self.n = 2

    def test_crude_age_standardization(self):
        crude = sm.crude_age_standardization(self.e, self.b, self.n).round(8)
        crude_exp = np.array([0.02375000, 0.02666667])
        assert list(crude) == list(crude_exp)

    def test_direct_age_standardization(self):
        direct = np.array(
            sm.direct_age_standardization(self.e, self.b, self.s_b, self.n)
        ).round(8)
        direct_exp = np.array(
            [[0.02374402, 0.01920491, 0.02904848], [0.02665072, 0.02177143, 0.03230508]]
        )
        assert list(direct.flatten()) == list(direct_exp.flatten())

    def test_indirect_age_standardization(self):
        indirect = np.array(
            sm.indirect_age_standardization(self.e, self.b, self.s_e, self.s_b, self.n)
        ).round(8)
        indirect_exp = np.array(
            [[0.02372382, 0.01940230, 0.02900789], [0.02610803, 0.02154304, 0.03164035]]
        )
        assert list(indirect.flatten()) == list(indirect_exp.flatten())


class TestSRate:
    def setup_method(self):
        sids = libpysal.io.open(libpysal.examples.get_path("sids2.dbf"), "r")
        self.w = libpysal.io.open(libpysal.examples.get_path("sids2.gal"), "r").read()
        self.b, self.e = np.array(sids[:, 8]), np.array(sids[:, 9])
        self.er = [0.453433, 0.000000, 0.775871, 0.973810, 3.133190]
        self.eb = [0.0016973, 0.0017054, 0.0017731, 0.0020129, 0.0035349]
        self.sr = [0.0009922, 0.0012639, 0.0009740, 0.0007605, 0.0050154]
        self.smr = [0.00083622, 0.00109402, 0.00081567, 0.0, 0.0048209]
        self.smr_w = [0.00127146, 0.00127146, 0.0008433, 0.0, 0.0049889]
        self.smr2 = [0.00091659, 0.00087641, 0.00091073, 0.0, 0.00467633]
        self.s_ebr10 = np.array(
            [
                4.01485749e-05,
                3.62437513e-05,
                4.93034844e-05,
                5.09387329e-05,
                3.72735210e-05,
                3.69333797e-05,
                5.40245456e-05,
                2.99806055e-05,
                3.73034109e-05,
                3.47270722e-05,
            ]
        ).reshape(-1, 1)

        stl_ex = libpysal.examples.load_example("stl")
        self.stl = libpysal.io.open(stl_ex.get_path("stl_hom.csv"), "r")
        self.stl_e, self.stl_b = np.array(self.stl[:, 10]), np.array(self.stl[:, 13])
        self.stl_w = libpysal.io.open(stl_ex.get_path("stl.gal"), "r").read()
        if not self.stl_w.id_order_set:
            self.stl_w.id_order = list(range(1, len(self.stl) + 1))

        if not PANDAS_EXTINCT:
            self.df = libpysal.io.open(libpysal.examples.get_path("sids2.dbf")).to_df()
            self.ename = "SID74"
            self.bname = "BIR74"
            self.stl_df = libpysal.io.open(
                libpysal.examples.get_path("stl_hom.csv")
            ).to_df()
            self.stl_ename = "HC7984"
            self.stl_bname = "PO7984"

    def test_excess_risk(self):
        out_er = sm.Excess_Risk(self.e, self.b).r
        np.testing.assert_allclose(out_er[:5].flatten(), self.er, rtol=RTOL, atol=ATOL)

    @pytest.mark.skipif(PANDAS_EXTINCT, reason="missing pandas")
    def test_excess_risk_tabular(self):
        out_er = sm.Excess_Risk(self.df[self.ename], self.df[self.bname]).r
        np.testing.assert_allclose(out_er[:5].flatten(), self.er, rtol=RTOL, atol=ATOL)
        assert isinstance(out_er, np.ndarray)
        out_er = sm.Excess_Risk.by_col(self.df, self.ename, self.bname)
        outcol = f"{self.ename}-{self.bname}_excess_risk"
        outcol = out_er[outcol]
        np.testing.assert_allclose(outcol[:5], self.er, rtol=RTOL, atol=ATOL)
        assert isinstance(outcol.values, np.ndarray)

    def test_empirical_bayes(self):
        out_eb = sm.Empirical_Bayes(self.e, self.b).r
        np.testing.assert_allclose(out_eb[:5].flatten(), self.eb, rtol=RTOL, atol=ATOL)

    @pytest.mark.skipif(PANDAS_EXTINCT, reason="missing pandas")
    def test_empirical_bayes_tabular(self):
        out_eb = sm.Empirical_Bayes(self.df[self.ename], self.df[self.bname]).r
        np.testing.assert_allclose(out_eb[:5].flatten(), self.eb, rtol=RTOL, atol=ATOL)
        assert isinstance(out_eb, np.ndarray)

        out_eb = sm.Empirical_Bayes.by_col(self.df, self.ename, self.bname)
        outcol = f"{self.ename}-{self.bname}_empirical_bayes"
        outcol = out_eb[outcol]
        np.testing.assert_allclose(outcol[:5], self.eb, rtol=RTOL, atol=ATOL)
        assert isinstance(outcol.values, np.ndarray)

    def test_spatial_empirical_bayes(self):
        s_eb = sm.Spatial_Empirical_Bayes(self.stl_e, self.stl_b, self.stl_w)
        np.testing.assert_allclose(self.s_ebr10, s_eb.r[:10], rtol=RTOL, atol=ATOL)

    @pytest.mark.skipif(PANDAS_EXTINCT, reason="missing pandas")
    def test_spatial_empirical_bayes_tabular(self):
        s_eb = sm.Spatial_Empirical_Bayes(
            self.stl_df[self.stl_ename], self.stl_df[self.stl_bname], self.stl_w
        ).r
        assert isinstance(s_eb, np.ndarray)
        np.testing.assert_allclose(self.s_ebr10, s_eb[:10])

        s_eb = sm.Spatial_Empirical_Bayes.by_col(
            self.stl_df, self.stl_ename, self.stl_bname, self.stl_w
        )
        outcol = f"{self.stl_ename}-{self.stl_bname}_spatial_empirical_bayes"
        r = s_eb[outcol].values
        assert isinstance(r, np.ndarray)
        np.testing.assert_allclose(self.s_ebr10, r[:10].reshape(-1, 1))

    def test_spatial_rate(self):
        out_sr = sm.Spatial_Rate(self.e, self.b, self.w).r
        np.testing.assert_allclose(out_sr[:5].flatten(), self.sr, rtol=RTOL, atol=ATOL)

    @pytest.mark.skipif(PANDAS_EXTINCT, reason="missing pandas")
    def test_spatial_rate_tabular(self):
        out_sr = sm.Spatial_Rate(self.df[self.ename], self.df[self.bname], self.w).r
        np.testing.assert_allclose(out_sr[:5].flatten(), self.sr, rtol=RTOL, atol=ATOL)
        assert isinstance(out_sr, np.ndarray)

        out_sr = sm.Spatial_Rate.by_col(self.df, self.ename, self.bname, w=self.w)
        outcol = f"{self.ename}-{self.bname}_spatial_rate"
        r = out_sr[outcol].values
        assert isinstance(r, np.ndarray)
        np.testing.assert_allclose(r[:5], self.sr, rtol=RTOL, atol=ATOL)

    def test_spatial_median_rate(self):
        out_smr = sm.Spatial_Median_Rate(self.e, self.b, self.w).r
        out_smr_w = sm.Spatial_Median_Rate(self.e, self.b, self.w, aw=self.b).r
        out_smr2 = sm.Spatial_Median_Rate(self.e, self.b, self.w, iteration=2).r
        np.testing.assert_allclose(
            out_smr[:5].flatten(), self.smr, atol=ATOL, rtol=RTOL
        )
        np.testing.assert_allclose(
            out_smr_w[:5].flatten(), self.smr_w, atol=ATOL, rtol=RTOL
        )
        np.testing.assert_allclose(
            out_smr2[:5].flatten(), self.smr2, atol=ATOL, rtol=RTOL
        )

    @pytest.mark.skipif(PANDAS_EXTINCT, reason="missing pandas")
    def test_spatial_median_rate_tabular(self):
        out_smr = sm.Spatial_Median_Rate(
            self.df[self.ename], self.df[self.bname], self.w
        ).r
        out_smr_w = sm.Spatial_Median_Rate(
            self.df[self.ename], self.df[self.bname], self.w, aw=self.df[self.bname]
        ).r
        out_smr2 = sm.Spatial_Median_Rate(
            self.df[self.ename], self.df[self.bname], self.w, iteration=2
        ).r

        assert isinstance(out_smr, np.ndarray)
        assert isinstance(out_smr_w, np.ndarray)
        assert isinstance(out_smr2, np.ndarray)

        np.testing.assert_allclose(
            out_smr[:5].flatten(), self.smr, atol=ATOL, rtol=RTOL
        )
        np.testing.assert_allclose(
            out_smr_w[:5].flatten(), self.smr_w, atol=ATOL, rtol=RTOL
        )
        np.testing.assert_allclose(
            out_smr2[:5].flatten(), self.smr2, atol=ATOL, rtol=RTOL
        )

        out_smr = sm.Spatial_Median_Rate.by_col(self.df, self.ename, self.bname, self.w)
        out_smr_w = sm.Spatial_Median_Rate.by_col(
            self.df, self.ename, self.bname, self.w, aw=self.df[self.bname]
        )
        out_smr2 = sm.Spatial_Median_Rate.by_col(
            self.df, self.ename, self.bname, self.w, iteration=2
        )
        outcol = f"{self.ename}-{self.bname}_spatial_median_rate"

        np.testing.assert_allclose(
            out_smr[outcol].values[:5], self.smr, rtol=RTOL, atol=ATOL
        )
        np.testing.assert_allclose(
            out_smr_w[outcol].values[:5], self.smr_w, rtol=RTOL, atol=ATOL
        )
        np.testing.assert_allclose(
            out_smr2[outcol].values[:5], self.smr2, rtol=RTOL, atol=ATOL
        )

    @pytest.mark.skipif(PANDAS_EXTINCT, reason="missing pandas")
    def test_Spatial_Smoother_multicol(self):
        # test that specifying multiple columns works correctly. Since the
        # function is shared over all spatial smoothers, we can only test one.
        enames = [self.ename, "SID79"]
        bnames = [self.bname, "BIR79"]
        out_df = sm.Spatial_Median_Rate.by_col(self.df, enames, bnames, self.w)
        outcols = [
            "{}-{}_spatial_median_rate".format(e, b) for e, b in zip(enames, bnames)
        ]
        smr79 = np.array(
            [0.00122129, 0.00176924, 0.00176924, 0.00240964, 0.00272035]
        )
        answers = [self.smr, smr79]
        for col, answer in zip(outcols, answers):
            np.testing.assert_allclose(
                out_df[col].values[:5], answer, rtol=1e-4, atol=1e-6
            )

    @pytest.mark.skipif(PANDAS_EXTINCT, reason="missing pandas")
    def test_Smoother_multicol(self):
        # test that non-spatial smoothers work with multicolumn queries
        enames = [self.ename, "SID79"]
        bnames = [self.bname, "BIR79"]
        out_df = sm.Excess_Risk.by_col(self.df, enames, bnames)
        outcols = ["{}-{}_excess_risk".format(e, b) for e, b in zip(enames, bnames)]
        er79 = np.array([0.000000, 2.796607, 0.8383863, 1.217479, 0.943811])
        answers = [self.er, er79]
        for col, answer in zip(outcols, answers):
            np.testing.assert_allclose(
                out_df[col].values[:5], answer, rtol=1e-4, atol=1e-7
            )


class TestHB:
    def setup_method(self):
        sids = libpysal.io.open(libpysal.examples.get_path("sids2.shp"), "r")
        self.sids = sids
        self.d = np.array([i.centroid for i in sids])
        self.w = KNN.from_array(self.d, k=5)
        if not self.w.id_order_set:
            self.w.id_order = self.w.id_order
        sids_db = libpysal.io.open(libpysal.examples.get_path("sids2.dbf"), "r")
        self.b, self.e = np.array(sids_db[:, 8]), np.array(sids_db[:, 9])
        self.sids_hb_rr5 = np.array([0.00075586, 0.0, 0.0008285, 0.0018315, 0.00498891])
        self.sids_hb_r2r5 = np.array(
            [0.0008285, 0.00084331, 0.00086896, 0.0018315, 0.00498891]
        )
        self.sids_hb_r3r5 = np.array(
            [0.00091659, 0.0, 0.00156838, 0.0018315, 0.00498891]
        )
        if not PANDAS_EXTINCT:
            self.df = sids_db.to_df()
            self.ename = "SID74"
            self.bname = "BIR74"
            self.enames = [self.ename, "SID79"]
            self.bnames = [self.bname, "BIR79"]
            self.sids79r = np.array([0.000563, 0.001659, 0.001879, 0.002410, 0.002720])

    @pytest.mark.skip("Deprecated")
    def test_triples(self):
        ht = sm.Headbanging_Triples(self.d, self.w)
        assert len(ht.triples) == len(self.d)
        ht2 = sm.Headbanging_Triples(self.d, self.w, edgecor=True)
        assert hasattr(ht2, "extra")
        assert len(ht2.triples) == len(self.d)
        htr = sm.Headbanging_Median_Rate(self.e, self.b, ht2, iteration=5)
        assert len(htr.r) == len(self.e)
        for i in htr.r:
            assert i is not None

    @pytest.mark.skip("Deprecated")
    def test_median_rate(self):
        s_ht = sm.Headbanging_Triples(self.d, self.w, k=5)
        sids_hb_r = sm.Headbanging_Median_Rate(self.e, self.b, s_ht)
        np.testing.assert_array_almost_equal(self.sids_hb_rr5, sids_hb_r.r[:5])
        sids_hb_r2 = sm.Headbanging_Median_Rate(self.e, self.b, s_ht, iteration=5)
        np.testing.assert_array_almost_equal(self.sids_hb_r2r5, sids_hb_r2.r[:5])
        sids_hb_r3 = sm.Headbanging_Median_Rate(self.e, self.b, s_ht, aw=self.b)
        np.testing.assert_array_almost_equal(self.sids_hb_r3r5, sids_hb_r3.r[:5])

    @pytest.mark.skip("Deprecated")
    def test_median_rate_tabular(self):
        # test that dataframe columns are treated correctly
        s_ht = sm.Headbanging_Triples(self.d, self.w, k=5)
        sids_hb_r = sm.Headbanging_Median_Rate(
            self.df[self.ename], self.df[self.bname], s_ht
        )
        self.assertIsInstance(sids_hb_r.r, np.ndarray)
        np.testing.assert_array_almost_equal(self.sids_hb_rr5, sids_hb_r.r[:5])

        sids_hb_r2 = sm.Headbanging_Median_Rate(
            self.df[self.ename], self.df[self.bname], s_ht, iteration=5
        )
        assert isinstance(sids_hb_r2.r, np.ndarray)
        np.testing.assert_array_almost_equal(self.sids_hb_r2r5, sids_hb_r2.r[:5])

        sids_hb_r3 = sm.Headbanging_Median_Rate(
            self.df[self.ename], self.df[self.bname], s_ht, aw=self.df[self.bname]
        )
        assert isinstance(sids_hb_r3.r, np.ndarray)
        np.testing.assert_array_almost_equal(self.sids_hb_r3r5, sids_hb_r3.r[:5])

        # test that the by col on multiple names works correctly
        sids_hr_df = sm.Headbanging_Median_Rate.by_col(
            self.df, self.enames, self.bnames, w=self.w
        )
        outcols = [
            f"{e}-{b}_headbanging_median_rate"
            for e, b in zip(self.enames, self.bnames, strict=True)
        ]
        for col, answer in zip(outcols, [self.sids_hb_rr5, self.sids79r], strict=True):
            this_col = sids_hr_df[col].values
            assert isinstance(this_col, np.ndarray)
            np.testing.assert_allclose(
                sids_hr_df[col][:5], answer, rtol=RTOL, atol=ATOL * 10
            )


class TestKernelAgeAdjSM:
    def setup_method(self):
        self.e = np.array([10, 1, 3, 4, 2, 5])
        self.b = np.array([100, 15, 20, 20, 80, 90])
        self.e1 = np.array([10, 8, 1, 4, 3, 5, 4, 3, 2, 1, 5, 3])
        self.b1 = np.array([100, 90, 15, 30, 25, 20, 30, 20, 80, 80, 90, 60])
        self.s = np.array([98, 88, 15, 29, 20, 23, 33, 25, 76, 80, 89, 66])
        self.points = [(10, 10), (20, 10), (40, 10), (15, 20), (30, 20), (30, 30)]
        self.kw = Kernel(self.points)
        self.points1 = np.array(self.points) + 5
        self.points1 = np.vstack((self.points1, self.points))
        self.kw1 = Kernel(self.points1)
        if not self.kw.id_order_set:
            self.kw.id_order = list(range(0, len(self.points)))
        if not PANDAS_EXTINCT:
            import pandas as pd

            dfa = np.array([self.e, self.b]).T
            dfb = np.array([self.e1, self.b1, self.s]).T
            self.dfa = pd.DataFrame(dfa, columns=["e", "b"])
            self.dfb = pd.DataFrame(dfb, columns=["e", "b", "s"])

        # answers
        self.kernel_exp = [
            0.10543301,
            0.0858573,
            0.08256196,
            0.09884584,
            0.04756872,
            0.04845298,
        ]
        self.ageadj_exp = [
            0.10519625,
            0.08494318,
            0.06440072,
            0.06898604,
            0.06952076,
            0.05020968,
        ]
        self.disk_exp = [
            0.12222222000000001,
            0.10833333,
            0.08055556,
            0.08944444,
            0.09944444,
            0.09351852,
        ]
        self.sf_exp = np.array([0.111111, 0.111111, 0.085106, 0.076923])

    def test_smoother(self):
        kr = sm.Kernel_Smoother(self.e, self.b, self.kw)
        np.testing.assert_allclose(kr.r.flatten(), self.kernel_exp)

    def test_age_adjusted_smoother(self):
        ar = sm.Age_Adjusted_Smoother(self.e1, self.b1, self.kw, self.s)
        np.testing.assert_allclose(ar.r, self.ageadj_exp)

    @pytest.mark.skipif(PANDAS_EXTINCT, reason="missing pandas")
    def test_age_adjusted_mmoother_tabular(self):
        dfb = self.dfb
        kr = sm.Age_Adjusted_Smoother(dfb.e, dfb.b, s=dfb.s, w=self.kw)
        assert isinstance(kr.r, np.ndarray)
        np.testing.assert_allclose(kr.r, self.ageadj_exp)

        kr = sm.Age_Adjusted_Smoother.by_col(dfb, "e", "b", w=self.kw, s="s")
        answer = np.array(
            [0.10519625, 0.08494318, 0.06440072, 0.06898604, 0.06952076, 0.05020968]
        )
        colname = "e-b_age_adjusted_smoother"
        np.testing.assert_allclose(kr[colname].values, answer, rtol=RTOL, atol=ATOL)

    def test_disk_smoother(self):
        self.kw.transform = "b"
        disk = sm.Disk_Smoother(self.e, self.b, self.kw)
        np.testing.assert_allclose(disk.r.flatten(), self.disk_exp)

    @pytest.mark.skipif(PANDAS_EXTINCT, reason="missing pandas")
    def test_disk_smoother_tabular(self):
        self.kw.transform = "b"
        dfa = self.dfa
        disk = sm.Disk_Smoother(dfa.e, dfa.b, self.kw).r
        np.testing.assert_allclose(disk.flatten(), self.disk_exp)

        disk = sm.Disk_Smoother.by_col(dfa, "e", "b", self.kw)
        col = "e-b_disk_smoother"
        np.testing.assert_allclose(
            disk[col].values, self.disk_exp, rtol=RTOL, atol=ATOL
        )

    def test_spatial_filtering(self):
        points = np.array(self.points)
        bbox = [[0, 0], [45, 45]]
        sf = sm.Spatial_Filtering(bbox, points, self.e, self.b, 2, 2, r=30)
        np.testing.assert_allclose(sf.r, self.sf_exp, rtol=RTOL, atol=ATOL)

    @pytest.mark.skipif(PANDAS_EXTINCT, reason="missing pandas")
    def test_spatial_filtering_tabular(self):
        point_array = np.array(self.points)
        bbox = [[0, 0], [45, 45]]
        dfa = self.dfa
        sf = sm.Spatial_Filtering(bbox, point_array, dfa.e, dfa.b, 2, 2, r=30)

        np.testing.assert_allclose(sf.r, self.sf_exp, rtol=RTOL, atol=ATOL)

        dfa["geometry"] = self.points
        sf = sm.Spatial_Filtering.by_col(dfa, "e", "b", 3, 3, r=30)
        r_answer = np.array(
            [
                0.07692308,
                0.07213115,
                0.07213115,
                0.07692308,
                0.07692308,
                0.07692308,
                0.07692308,
                0.07692308,
                0.07692308,
            ]
        )
        x_answer = np.array([10.0, 10.0, 10.0, 20.0, 20.0, 20.0, 30.0, 30.0, 30.0])
        y_answer = np.array(
            [
                10.000000,
                16.666667,
                23.333333,
                10.000000,
                16.666667,
                23.333333,
                10.000000,
                16.666667,
                23.333333,
            ]
        )
        columns = [f"e-b_spatial_filtering_{name}" for name in ["X", "Y", "R"]]

        for col, answer in zip(columns, [x_answer, y_answer, r_answer], strict=True):
            np.testing.assert_allclose(sf[col].values, answer, rtol=RTOL, atol=ATOL)


class TestUtils:
    def test_sum_by_n(self):
        d = np.array([10, 9, 20, 30])
        w = np.array([0.5, 0.1, 0.3, 0.8])
        n = 2
        exp_sum = np.array([5.9, 30.0])
        np.testing.assert_array_almost_equal(exp_sum, sm.sum_by_n(d, w, n))

    def test_standardized_mortality_ratio(self):
        e = np.array([30, 25, 25, 15, 33, 21, 30, 20])
        b = np.array([100, 100, 110, 90, 100, 90, 110, 90])
        s_e = np.array([100, 45, 120, 100, 50, 30, 200, 80])
        s_b = np.array([1000, 900, 1000, 900, 1000, 900, 1000, 900])
        n = 2
        exp_smr = np.array([2.48691099, 2.73684211])
        np.testing.assert_array_almost_equal(
            exp_smr, sm.standardized_mortality_ratio(e, b, s_e, s_b, n)
        )

    def test_choynowski(self):
        e = np.array([30, 25, 25, 15, 33, 21, 30, 20])
        b = np.array([100, 100, 110, 90, 100, 90, 110, 90])
        n = 2
        exp_choy = np.array([0.30437751, 0.29367033])
        np.testing.assert_array_almost_equal(exp_choy, sm.choynowski(e, b, n))

    def test_assuncao_rate(self):
        e = np.array([30, 25, 25, 15, 33, 21, 30, 20])
        b = np.array([100, 100, 110, 90, 100, 90, 110, 90])
        exp_assuncao = np.array([1.03843594, -0.04099089, -0.56250375, -1.73061861])
        np.testing.assert_array_almost_equal(exp_assuncao, sm.assuncao_rate(e, b)[:4])
